from gatox.models.runner import Runner
from gatox.models.repository import Repository
from gatox.models.secret import Secret


class Organization:
    """ """

    def __init__(self, org_data: dict, user_scopes: list, limited_data: bool = False):
        """Wrapper object for an organization.

        Args:
            org_data (dict): Org data from GitHub API
            user_scopes (list): List of OAuth scopes that the PAT has
            limited_data (bool): Whether limited org_data is present (default: False)
        """
        self.name = None
        self.org_admin_user = False
        self.org_admin_scopes = False
        self.org_member = False
        self.secrets: list[Secret] = []
        self.runners: list[Runner] = []
        self.sso_enabled = False

        self.limited_data = limited_data

        self.public_repos = []
        self.private_repos = []

        self.name = org_data["login"]

        # If fields such as billing email are populated, then the user MUST
        # be an organization owner. If not, then the user is a member (for
        # private repos) or
        if "billing_email" in org_data and org_data["billing_email"] is not None:
            if "admin:org" in user_scopes:
                self.org_admin_scopes = True
            self.org_admin_user = True
            self.org_member = True
        elif "billing_email" in org_data:
            self.org_admin_user = False
            self.org_member = True
        else:
            self.org_admin_user = False
            self.org_member = False

    def set_secrets(self, secrets: list[Secret]):
        """Set repo-level secrets.

        Args:
            secrets (list): List of secrets at the organization level.
        """
        self.secrets = secrets

    def set_repository(self, repo: Repository):
        """Add a single repository to the organization.

        Args:
            repo (Repository): Single repository object.
        """
        if repo.is_private():
            self.private_repos.append(repo)
        else:
            self.public_repos.append(repo)

    def set_public_repos(self, repos: list[Repository]):
        """List of public repos for the org.

        Args:
            repos (List[Repository]): List of Repository wrapper objects.
        """
        self.public_repos = repos

    def set_private_repos(self, repos: list[Repository]):
        """List of private repos for the org.

        Args:
            repos (List[Repository]): List of Repository wrapper objects.
        """
        self.private_repos = repos

    def set_runners(self, runners: list[Runner]):
        """Set a list of runners that the organization can access.

        Args:
            runners (List[Runner]): List of runners that are attached to the
            organization.
        """
        self.runners = runners

    def toJSON(self):
        """Converts the repository to a Gato JSON representation."""
        if self.limited_data:
            representation = {"name": self.name}
        else:
            representation = {
                "name": self.name,
                "org_admin_user": self.org_admin_user,
                "org_member": self.org_member,
                "org_runners": [runner.toJSON() for runner in self.runners],
                "org_secrets": [secret.toJSON() for secret in self.secrets],
                "sso_access": self.sso_enabled,
                "public_repos": [
                    repository.toJSON() for repository in self.public_repos
                ],
                "private_repos": [
                    repository.toJSON() for repository in self.private_repos
                ],
            }

        return representation



"""
Copyright 2025, Adnan Khan

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from gatox.workflow_graph.nodes.node import Node
from gatox.workflow_graph.nodes.job import JobNode
from gatox.models.workflow import Workflow


class WorkflowNode(Node):
    """Workflow node"""

    def __init__(self, ref: str, repo_name: str, workflow_path: str):
        """Constructor for workflow wrapper."""

        # Create a unique ID for this workflow.
        self.name = f"{repo_name}:{ref}:{workflow_path}"
        # By default, a workflow node is "uninitialized" until it is processed
        # with the workflow YAML. We sometimes add unititialized nodes to the
        # graph if a workflow references another workflow that has not been
        # processed yet.
        self.uninitialized = True
        self.__workflow_path = workflow_path
        self.__triggers = []
        self.__callers = []
        self.__excluded = False
        self.__repo_name = repo_name
        self.__env_vars = {}
        self.inputs = {}

    def __hash__(self):
        return hash((self.name, self.__class__.__name__))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.name == other.name

    def set_params(self, params):
        """ """
        self.params = params

    def get_parts(self):
        """ """
        repo, ref, path = self.name.split(":")

        return repo, ref, path

    def repo_name(self):
        """ """
        return self.__repo_name

    def get_workflow_name(self):
        """
        Get name of the workflow file associated with the JobNode instance.

        Returns:
            str: The path to the workflow file.
        """
        return self.__workflow_path.replace(".github/workflows/", "")

    def __process_triggers(self, workflow_data: dict):
        """Retrieve the triggers associated with the Workflow node."""
        triggers = workflow_data.get("on", [])
        extracted_triggers = []

        if isinstance(triggers, list):
            return triggers
        elif isinstance(triggers, str):
            return [triggers]
        elif isinstance(triggers, dict):
            for trigger, trigger_conditions in triggers.items():
                if trigger == "pull_request_target":
                    if trigger_conditions and "types" in trigger_conditions:
                        if (
                            "labeled" in trigger_conditions["types"]
                            and len(trigger_conditions["types"]) == 1
                        ):
                            extracted_triggers.append(
                                f"{trigger}:{trigger_conditions['types'][0]}"
                            )
                        else:
                            extracted_triggers.append(trigger)
                    else:
                        extracted_triggers.append(trigger)
                elif trigger == "workflow_run":
                    if "branches" in trigger_conditions:
                        # If the branches filter is present, then
                        # forks cannot trigger the workflow.
                        self.__excluded = True
                    else:
                        extracted_triggers.append(trigger)
                else:
                    extracted_triggers.append(trigger)

        return extracted_triggers

    def __process_inputs(self, workflow_data: dict):
        try:
            if (
                "workflow_dispatch" in self.__triggers
                and isinstance(workflow_data["on"], dict)
                and isinstance(workflow_data["on"]["workflow_dispatch"], dict)
                and "inputs" in workflow_data["on"]["workflow_dispatch"]
            ):
                return workflow_data["on"]["workflow_dispatch"]["inputs"]
            else:
                return {}
        except TypeError:
            print(workflow_data["on"])

    def __process_envs(self, workflow_data: dict):
        if "env" in workflow_data:
            return workflow_data["env"]
        else:
            return {}

    def excluded(self):
        """Returns whether the workflow is excluded as it cannot be
        triggered from a fork."""
        return self.__excluded

    def get_env_vars(self):
        """Returns environment variables for the workflow."""
        return self.__env_vars

    def add_caller_reference(self, caller: JobNode):
        """Add a reference to a JobNode that calls this Workflow node,
        if it is not already marked (as we can reach it multiple times
        for nested relationships).
        """
        if caller not in self.__callers:
            self.__callers.append(caller)

    def get_caller_workflows(self):
        """Retrieve a set of the workflows that call this Workflow node."""
        if not self.__callers:
            return set()
        else:
            return set([caller.get_workflow() for caller in self.__callers])

    def initialize(self, workflow: Workflow):
        """Initialize the Workflow node with the parsed workflow data."""
        self.__triggers = self.__process_triggers(workflow.parsed_yml)
        self.__env_vars = self.__process_envs(workflow.parsed_yml)

        self.inputs = self.__process_inputs(workflow.parsed_yml)
        self.uninitialized = False

    def get_triggers(self):
        """Retrieve the triggers associated with the Workflow node."""
        return self.__triggers

    def get_tags(self):
        """ """
        tags = set([self.__class__.__name__])

        if self.uninitialized:
            tags.add("uninitialized")
        else:
            tags.add("initialized")

        for trigger in self.__triggers:
            tags.add(trigger)

        return tags

    def get_attrs(self):
        """Retrieve node attributes associated with the Workflow node."""
        return {}



import pytest
from unittest.mock import Mock, patch
from gatox.workflow_graph.graph.tagged_graph import TaggedGraph
from gatox.workflow_graph.visitors.injection_visitor import InjectionVisitor
from gatox.github.api import Api
from gatox.workflow_graph.nodes.workflow import WorkflowNode
from gatox.workflow_graph.nodes.job import JobNode
from gatox.workflow_graph.nodes.step import StepNode
from gatox.workflow_graph.nodes.action import ActionNode


@pytest.fixture
def mock_api():
    mock_api = Mock(spec=Api)
    mock_api.get_all_environment_protection_rules.return_value = {}
    return mock_api


@pytest.fixture
def mock_graph():
    return Mock(spec=TaggedGraph)


def test_find_injections_no_nodes(mock_graph, mock_api, capsys):
    """Test when no nodes are found with injection tags"""
    mock_graph.get_nodes_for_tags.return_value = []

    results = InjectionVisitor.find_injections(mock_graph, mock_api)
    assert results == {}
    # captured = capsys.readouterr()
    # assert "INJECT:" in captured.out
    # mock_graph.get_nodes_for_tags.assert_called_once()


# def test_find_injections_with_workflow_run(mock_graph, mock_api):
#     """Test that workflow_run tag is included when ignore_workflow_run is False"""
#     InjectionVisitor.find_injections(mock_graph, mock_api, ignore_workflow_run=False)

#     call_args = mock_graph.get_nodes_for_tags.call_args[0][0]
#     assert "workflow_run" in call_args


# def test_find_injections_ignore_workflow_run(mock_graph, mock_api):
#     """Test that workflow_run tag is excluded when ignore_workflow_run is True"""
#     InjectionVisitor.find_injections(mock_graph, mock_api, ignore_workflow_run=True)

#     call_args = mock_graph.get_nodes_for_tags.call_args[0][0]
#     assert "workflow_run" not in call_args


# def test_find_injections_with_path(mock_graph, mock_api, capsys):
#     """Test path analysis with various node types"""
#     workflow_node = Mock(spec=WorkflowNode)
#     workflow_node.get_tags.return_value = ["WorkflowNode"]
#     workflow_node.get_env_vars.return_value = {"TEST_ENV": "github.event.test"}

#     job_node = Mock(spec=JobNode)
#     job_node.get_tags.return_value = ["JobNode"]
#     job_node.get_env_vars.return_value = {}
#     job_node.deployments = []

#     step_node = Mock(spec=StepNode)
#     step_node.get_tags.return_value = ["StepNode", "injectable"]
#     step_node.contexts = ["github.event.test"]

#     path = [workflow_node, job_node, step_node]

#     mock_graph.get_nodes_for_tags.return_value = [workflow_node]
#     mock_graph.dfs_to_tag.return_value = [path]

#     InjectionVisitor.find_injections(mock_graph, mock_api)

#     captured = capsys.readouterr()
#     assert "INJECT:" in captured.out


# def test_find_injections_with_approval_gate(mock_graph, mock_api):
#     """Test path analysis with approval gate"""
#     job_node = Mock(spec=JobNode)
#     job_node.get_tags.return_value = ["JobNode"]
#     job_node.__repo_name = "test/repo"
#     job_node.deployments = ["prod"]
#     job_node.get_env_vars.return_value = {}

#     step_node = Mock(spec=StepNode)
#     step_node.get_tags.return_value = ["StepNode", "injectable"]
#     step_node.contexts = []

#     path = [job_node, step_node]

#     mock_graph.get_nodes_for_tags.return_value = [job_node]
#     mock_graph.dfs_to_tag.return_value = [path]
#     mock_api.get_all_environment_protection_rules.return_value = {"prod": True}

#     InjectionVisitor.find_injections(mock_graph, mock_api)

#     mock_api.get_all_environment_protection_rules.assert_called_with("test/repo")


def test_find_injections_with_action_node(mock_graph, mock_api):
    """Test path analysis with action node"""
    action_node = Mock(spec=ActionNode)
    action_node.get_tags.return_value = ["ActionNode"]

    mock_graph.get_nodes_for_tags.return_value = [action_node]
    mock_graph.dfs_to_tag.return_value = [[action_node]]

    InjectionVisitor.find_injections(mock_graph, mock_api)

    mock_graph.get_nodes_for_tags.assert_called_once()



import logging
import requests
import json

from gatox.cli.output import Output
from gatox.configuration.configuration_manager import ConfigurationManager
from gatox.github.search import Search
from gatox.github.api import Api


logger = logging.getLogger(__name__)


class Searcher:
    """Class that encapsulates functionality to use the GitHub code search API."""

    def __init__(
        self,
        pat: str,
        socks_proxy: str = None,
        http_proxy: str = None,
        github_url: str = None,
    ):

        self.api = Api(
            pat,
            socks_proxy=socks_proxy,
            http_proxy=http_proxy,
            github_url=github_url,
        )

        if self.api.proxies:
            self.proxies = self.api.proxies
        else:
            self.proxies = None

        self.socks_proxy = socks_proxy
        self.http_proxy = http_proxy
        self.user_perms = None

    def __setup_user_info(self):
        """Checks the PAT to ensure that it is valid and retrieves the
        associated scopes.

        Returns:
            bool: If the PAT is associated with a valid user.
        """
        if not self.user_perms:
            self.user_perms = self.api.check_user()
            if not self.user_perms:
                Output.error("This token cannot be used for enumeration!")
                return False

            Output.info(
                f"The authenticated user is: "
                f"{Output.bright(self.user_perms['user'])}"
            )
            if len(self.user_perms["scopes"]) > 0:
                Output.info(
                    f"The GitHub Classic PAT has the following scopes: "
                    f'{Output.yellow(", ".join(self.user_perms["scopes"]))}'
                )
            else:
                Output.warn("The token has no scopes!")

        return True

    def use_sourcegraph_api(self, organization: str, query=None, output_text=None):
        """
        This method is used to search for repositories in an organization using the Sourcegraph API.
        It constructs a search query and sends a GET request to the Sourcegraph search API.
        The results are streamed and added to a set.

        Args:
            organization (str): The name of the organization to search in.
            query (str, optional): A custom search query. If not provided, a default query is used.

        Returns:
            set: A set of search results.
        """
        repo_filter = f"repo:{organization}/ " if organization else ""
        url = "https://sourcegraph.com/.api/search/stream"
        headers = {"Content-Type": "application/json"}
        params = {
            "q": (
                "context:global "
                "self-hosted OR "
                "(runs-on AND NOT "
                f"/({'|'.join(ConfigurationManager().WORKFLOW_PARSING['GITHUB_HOSTED_LABELS'])})/) "
                f"{repo_filter}"
                "lang:YAML file:.github/workflows/ count:100000"
            )
        }

        if query:
            Output.info(
                f"Searching SourceGraph with the following query: {Output.bright(query)}"
            )
            params["q"] = query
        else:
            Output.info(
                f"Searching SourceGraph with the default Gato query: {Output.bright(params['q'])}"
            )
        response = requests.get(
            url, headers=headers, params=params, stream=True, proxies=self.proxies
        )
        results = set()
        if response.status_code == 200:
            for line in response.iter_lines():
                if line and line.decode().startswith("data:"):
                    json_line = line.decode().replace("data:", "").strip()
                    event = json.loads(json_line)

                    if "title" in event and event["title"] == "Unable To Process Query":
                        Output.error("SourceGraph was unable to process the query!")
                        Output.error(f"Error: {Output.bright(event['description'])}")
                        return False

                    for element in event:
                        if "repository" in element:
                            results.add(
                                element["repository"].replace("github.com/", "")
                            )
        else:
            Output.error(
                f"SourceGraph returned an error: {Output.bright(response.status_code)}"
            )
            return False

        return sorted(results)

    def use_search_api(self, organization: str, query=None):
        """Utilize GitHub Code Search API to try and identify repositories
        using self-hosted runners. This is subject to a high false-positive
        rate because any occurance of 'self-hosted' within a YAML file will
        yield a positive result. This is ideally used as a first line method to
        retrieve a list of candidate repos that can then be enumerated using
        methods like YAML parsing and action log analysis.

        Args:
            organization (str): Organization to enumerate using
            the GitHub code search API.
            query (str, optional): Custom code-search query.

        Returns:
            list: List of repositories suspected of using self-hosted runners
            as identified by GitHub code search.
        """
        self.__setup_user_info()

        if not self.user_perms:
            return False

        api_search = Search(self.api)

        if query:
            Output.info(
                f"Searching GitHub with the following query: {Output.bright(query)}"
            )
        else:
            Output.info(
                f"Searching repositories within {Output.bright(organization)} "
                "using the GitHub Code Search API for 'self-hosted' within "
                "YAML files."
            )
        candidates = api_search.search_enumeration(organization, custom_query=query)

        return sorted(candidates)

    def present_results(self, results, output_text=None):
        """
        This method is used to present the results of the search. It first
        prints the number of non-fork repositories that matched the criteria.
        If an output_text file path is provided, it writes the results into
        that file. Finally, it prints each result in a tabbed format.

        Args:
            results (list): A list of non-fork repositories that matched the
            criteria.
            output_text (str, optional): The file path where the results
            should be written. Defaults to None.
        """
        Output.result(
            f"Identified {len(results)} non-fork repositories that matched "
            "the criteria!"
        )

        if output_text:
            with open(output_text, "w") as file_output:
                for candidate in results:
                    file_output.write(f"{candidate}\n")

        for candidate in results:
            Output.tabbed(candidate)



"""
Copyright 2025, Adnan Khan

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from gatox.workflow_graph.nodes.node import Node

from gatox.workflow_parser.utility import (
    process_matrix,
    process_runner,
    validate_if_check,
)


class JobNode(Node):
    """
    Wrapper class for a GitHub Actions workflow job.

    Attributes:
        name (str): A unique identifier for the job node.
        params (dict): Parameters associated with the job node.
    """

    def __init__(self, job_name: str, ref: str, repo_name: str, workflow_path: str):
        """
        Constructor for the job wrapper.

        Args:
            job_name (str): The name of the job.
            ref (str): The reference (e.g., branch or tag).
            repo_name (str): The name of the repository.
            workflow_path (str): The path to the workflow file.
        """
        # Create a unique ID for this step.
        self.name = f"{repo_name}:{ref}:{workflow_path}:{job_name}"
        self.ref = ref
        self.__workflow_path = workflow_path
        self.params = {}
        self.__repo_name = repo_name
        self.if_condition = None
        self.if_evaluation = None
        self.wf_reference = None
        self.needs = []
        self.deployments = []
        self.__env_vars = {}
        self.self_hosted = False
        self.outputs = {}

    def __hash__(self):
        """
        Return the hash value of the JobNode instance.

        Returns:
            int: The hash value of the JobNode instance.
        """
        return hash((self.name, self.__class__.__name__))

    def _check_selfhosted(self, job_def: dict):
        """Returns true if the job might run on a self-hosted runner."""

        # Easy
        if "self-hosted" in job_def["runs-on"]:
            return True
        # Process a matrix job
        elif "matrix." in job_def["runs-on"]:
            return process_matrix(job_def, job_def["runs-on"])
        # Process standard label
        else:
            return process_runner(job_def["runs-on"])

    def get_workflow_name(self):
        """
        Get name of the workflow file associated with the JobNode instance.

        Returns:
            str: The path to the workflow file.
        """
        return self.__workflow_path.replace(".github/workflows/", "")

    def get_workflow(self):
        return self.wf_reference

    def get_env_vars(self):
        """Returns environemnt variables used by the job in dictionary format."""
        return self.__env_vars

    def repo_name(self):
        """Return the repository name."""
        return self.__repo_name

    def get_if_eval(self):
        """ """
        return self.if_evaluation

    def evaluate_if(self):
        """ """
        if self.if_condition:
            return validate_if_check(self.if_condition, self.__env_vars)
        return True

    def populate(self, job_def, wf_node):

        self.wf_reference = wf_node

        if not isinstance(job_def, dict):
            raise ValueError(
                "Job definition is not a dictionary, the workflow yaml is likely invalid."
            )

        self.outputs = job_def.get("outputs", {})

        params = job_def.get("with", {})
        if params:
            self.set_params(params)

        if "runs-on" in job_def:
            self.self_hosted = self._check_selfhosted(job_def)

        if "environment" in job_def:
            if type(job_def["environment"]) == list:
                self.deployments.extend(job_def["environment"])
            else:
                self.deployments.append(job_def["environment"])

        if "env" in job_def and type(job_def["env"]) == dict:
            self.__env_vars = job_def["env"]

        if job_def and "if" in job_def:
            self.if_condition = job_def["if"]
            if type(self.if_condition) == str:
                self.if_condition = self.if_condition.replace("\n", "")
                self.if_evaluation = validate_if_check(self.if_condition, {})
            else:
                self.if_condition = None
        else:
            self.if_evaluation = True

    def __eq__(self, other):
        """
        Check if two JobNode instances are equal.

        Args:
            other (JobNode): Another JobNode instance to compare with.

        Returns:
            bool: True if the instances are equal, False otherwise.
        """
        return isinstance(other, self.__class__) and self.name == other.name

    def get_if(self):
        return self.if_condition

    def get_repr(self):
        """
        Get the representation of the Node instance.

        Returns:
            value: A dict representation of the Node instance.
        """

        value = {
            "node": str(self),
        }
        if self.get_if():
            value["if"] = self.get_if()
            if self.if_evaluation is not None and type(self.if_evaluation) is bool:
                value["if_eval"] = self.if_evaluation

        return value

    def set_params(self, params):
        self.params = params

    def get_needs(self):
        """ """
        return self.needs

    def add_needs(self, need_node):
        """
        Add a need to the JobNode instance.
        """
        self.needs.append(need_node)

    def get_tags(self):
        """
        Get the tags associated with the JobNode instance.

        Returns:
            set: A set containing the class name of the JobNode instance.
        """
        tags = set([self.__class__.__name__])

        if self.self_hosted:
            tags.add("self-hosted")

        return tags

    def get_attrs(self):
        """
        Get the attributes associated with the JobNode instance.

        Returns:
            dict: A dictionary containing attributes of the JobNode instance.
        """
        attr_dict = {
            self.__class__.__name__: True,
            "is_soft_gate": False,
            "is_hard_gate": False,
        }
        return attr_dict



import pytest
from unittest import mock
from unittest.mock import MagicMock, patch
from gatox.workflow_graph.visitors.pwn_request_visitor import PwnRequestVisitor
from gatox.workflow_graph.graph.tagged_graph import TaggedGraph
from gatox.github.api import Api
from gatox.workflow_graph.visitors.visitor_utils import VisitorUtils
from gatox.caching.cache_manager import CacheManager


@pytest.fixture
def mock_graph():
    return MagicMock(spec=TaggedGraph)


@pytest.fixture
def mock_api():
    return MagicMock(spec=Api)


@pytest.fixture
def mock_cache_manager():
    with patch(
        "gatox.workflow_graph.visitors.pwn_request_visitor.CacheManager"
    ) as MockCache:
        instance = MockCache.return_value
        instance.get_repository.return_value = MagicMock(is_fork=lambda: False)
        yield instance


def test_find_pwn_requests_no_nodes(mock_graph, mock_api, mock_cache_manager):
    mock_graph.get_nodes_for_tags.return_value = []
    with patch.object(VisitorUtils, "add_repo_results") as mock_add:
        result = PwnRequestVisitor.find_pwn_requests(mock_graph, mock_api)


def test_find_pwn_requests_with_nodes(mock_graph, mock_api, mock_cache_manager):
    node = MagicMock()
    mock_graph.get_nodes_for_tags.return_value = [node]
    mock_graph.dfs_to_tag.return_value = [[node]]

    with (
        patch.object(PwnRequestVisitor, "_process_single_path") as mock_process,
        patch.object(VisitorUtils, "add_repo_results") as mock_add,
    ):
        result = PwnRequestVisitor.find_pwn_requests(mock_graph, mock_api)
        mock_process.assert_called_once()



from typing import List
from multiprocessing import Process

from gatox.models.organization import Organization
from gatox.models.repository import Repository
from gatox.models.secret import Secret
from gatox.models.runner import Runner
from gatox.github.api import Api


class OrganizationEnum:
    """Helper class to wrap organization specific enumeration funcionality."""

    def __init__(self, api: Api):
        """Simple init method.

        Args:
            api (Api): Insantiated GitHub API wrapper object.
        """
        self.api = api

    def __assemble_repo_list(
        self, organization: str, visibilities: list
    ) -> List[Repository]:
        """Get a list of repositories that match the visibility types.

        Args:
            organization (str): Name of the organization.
            visibilities (list): List of visibilities (public, private, etc)
        """

        repos = []
        for visibility in visibilities:
            raw_repos = self.api.check_org_repos(organization, visibility)
            if raw_repos:
                repos.extend([Repository(repo) for repo in raw_repos])

        return repos

    def construct_repo_enum_list(self, organization: Organization) -> List[Repository]:
        """Constructs a list of repositories that a user has access to within
        an organization.

        Args:
            organization (Organization): Organization wrapper object.

        Returns:
            List[Repository]: List of repositories to enumerate.
        """
        org_private_repos = self.__assemble_repo_list(
            organization.name, ["private", "internal"]
        )

        # We might legitimately have no private repos despite being a
        # member.
        if org_private_repos:
            sso_enabled = self.api.validate_sso(
                organization.name, org_private_repos[0].name
            )
            organization.sso_enabled = sso_enabled
        else:
            org_private_repos = []

        org_public_repos = self.__assemble_repo_list(organization.name, ["public"])

        if organization.sso_enabled:
            return org_private_repos + org_public_repos
        else:
            return org_public_repos

    def admin_enum(self, organization: Organization):
        """Enumeration tasks to perform if the user is an org admin and the
        token has the necessary scopes.
        """
        if organization.org_admin_scopes and organization.org_admin_user:

            runners = self.api.check_org_runners(organization.name)
            if runners:
                org_runners = [
                    Runner(
                        runner["name"],
                        machine_name=None,
                        os=runner["os"],
                        status=runner["status"],
                        labels=runner["labels"],
                    )
                    for runner in runners["runners"]
                ]
                organization.set_runners(org_runners)

            org_secrets = self.api.get_org_secrets(organization.name)
            if org_secrets:
                org_secrets = [
                    Secret(secret, organization.name) for secret in org_secrets
                ]

                organization.set_secrets(org_secrets)



import unittest
from unittest.mock import Mock, patch

from gatox.enumerate.results.injection_result import InjectionResult
from gatox.enumerate.results.confidence import Confidence
from gatox.enumerate.results.complexity import Complexity


class TestInjectionResult(unittest.TestCase):
    def setUp(self):
        # Mock node objects for the attack path
        self.start_node = Mock()
        self.start_node.repo_name.return_value = "test/repo"
        self.start_node.get_workflow_name.return_value = "test.yml"
        self.start_node.get_triggers.return_value = [
            "pull_request_target",
            "push",
            "workflow_run",
        ]

        self.end_node = Mock()
        self.end_node.contexts = [
            "github.event.issue.body",
            "github.event.comment.body",
        ]

        self.path = [self.start_node, self.end_node]

        # Create InjectionResult instance
        self.result = InjectionResult(
            path=self.path,
            confidence_score=Confidence.HIGH,
            attack_complexity_score=Complexity.ZERO_CLICK,
        )

    def test_initialization(self):
        """Test proper initialization of InjectionResult"""
        self.assertEqual(self.result.repo_name(), "test/repo")
        self.assertEqual(self.result.issue_type(), "InjectionResult")
        self.assertEqual(self.result.confidence_score(), Confidence.HIGH)
        self.assertEqual(self.result.attack_complexity(), Complexity.ZERO_CLICK)

    def test_get_first_and_last_hash(self):
        """Test hash generation is consistent and includes required components"""
        first_hash = self.result.get_first_and_last_hash()
        second_hash = self.result.get_first_and_last_hash()

        # Same input should produce same hash
        self.assertEqual(first_hash, second_hash)

        # Different confidence should produce different hash
        different_result = InjectionResult(
            path=self.path,
            confidence_score=Confidence.LOW,
            attack_complexity_score=Complexity.ZERO_CLICK,
        )
        self.assertNotEqual(first_hash, different_result.get_first_and_last_hash())

    def test_filter_triggers(self):
        """Test trigger filtering logic"""
        triggers = ["push", "pull_request_target", "workflow_run", "schedule"]
        filtered = self.result.filter_triggers(triggers)

        # Should keep relevant triggers
        self.assertIn("pull_request_target", filtered)
        self.assertIn("workflow_run", filtered)

        # Should remove irrelevant triggers
        self.assertNotIn("push", filtered)
        self.assertNotIn("schedule", filtered)

    def test_to_machine(self):
        """Test machine-readable output format"""
        machine_output = self.result.to_machine()

        expected_keys = {
            "repository_name",
            "issue_type",
            "triggers",
            "initial_workflow",
            "confidence",
            "attack_complexity",
            "explanation",
            "path",
            "injectable_context",
        }

        # Verify all expected keys are present
        self.assertEqual(set(machine_output.keys()), expected_keys)

        # Verify content of machine output
        self.assertEqual(machine_output["repository_name"], "test/repo")
        self.assertEqual(machine_output["issue_type"], "InjectionResult")
        self.assertEqual(machine_output["initial_workflow"], "test.yml")
        self.assertEqual(
            machine_output["injectable_context"],
            ["github.event.issue.body", "github.event.comment.body"],
        )

        # Verify filtered triggers
        self.assertIn("pull_request_target", machine_output["triggers"])
        self.assertIn("workflow_run", machine_output["triggers"])
        self.assertNotIn("push", machine_output["triggers"])


if __name__ == "__main__":
    unittest.main()



from datetime import datetime
import yaml

from yaml import CSafeLoader
from yaml.resolver import Resolver


# remove resolver entries for On/Off/Yes/No
for ch in "OoTtFf":
    if len(Resolver.yaml_implicit_resolvers[ch]) == 1:
        del Resolver.yaml_implicit_resolvers[ch]
    else:
        Resolver.yaml_implicit_resolvers[ch] = [
            x
            for x in Resolver.yaml_implicit_resolvers[ch]
            if x[0] != "tag:yaml.org,2002:bool"
        ]


class Workflow:
    def __init__(
        self,
        repo_name,
        workflow_contents,
        workflow_name,
        default_branch="main",
        date=None,
        non_default=None,
        special_path=None,
    ):
        self.repo_name = repo_name
        self.invalid = False
        self.workflow_name = workflow_name
        self.special_path = special_path
        if non_default:
            self.branch = non_default
        else:
            self.branch = default_branch

        # Only save off if it's a valid parse. RAM matters.
        try:
            if type(workflow_contents) == bytes:
                workflow_contents = workflow_contents.decode("utf-8")
            self.parsed_yml = yaml.load(
                workflow_contents.replace("\t", "  "), Loader=CSafeLoader
            )

            if (
                "dependabot" in workflow_name
                and "- package-ecosystem:" in workflow_contents
            ):
                self.invalid = True

            if not self.parsed_yml or type(self.parsed_yml) != dict:
                self.invalid = True

            self.workflow_contents = workflow_contents
        except (
            yaml.parser.ParserError,
            yaml.scanner.ScannerError,
            yaml.constructor.ConstructorError,
        ) as parse_error:
            self.invalid = True
        except ValueError as parse_error:
            self.invalid = True
        except Exception as parse_error:
            print(
                "Received an exception while parsing workflow contents: "
                + str(parse_error)
            )
            self.invalid = True

        self.date = date if date else datetime.now().isoformat()

    def getPath(self):
        return f".github/workflows/{self.workflow_name}"

    def isInvalid(self):
        return self.invalid



import logging

from gatox.caching.cache_manager import CacheManager
from gatox.enumerate.reports.runners import RunnersReport
from gatox.cli.output import Output
from gatox.enumerate.reports.actions import ActionsReport
from gatox.models.execution import Repository
from gatox.models.secret import Secret
from gatox.models.runner import Runner
from gatox.github.api import Api

logger = logging.getLogger(__name__)


class RepositoryEnum:
    """Repository specific enumeration functionality."""

    def __init__(self, api: Api, skip_log: bool):
        """Initialize enumeration class with instantiated API wrapper and CLI
        parameters.

        Args:
            api (Api): GitHub API wraper object.
        """
        self.api = api
        self.skip_log = skip_log

    def perform_runlog_enumeration(self, repository: Repository, workflows: list):
        """Enumerate for the presence of a self-hosted runner based on
        downloading historical runlogs.

        Args:
            repository (Repository): Wrapped repository object.
            workflows (list): List of workflows that execute on self-hosted runner.

        Returns:
            bool: True if a self-hosted runner was detected.
        """
        runner_detected = False
        wf_runs = []

        wf_runs = self.api.retrieve_run_logs(repository.name, workflows=workflows)

        if wf_runs:
            for wf_run in wf_runs:
                runner = Runner(
                    wf_run["runner_name"],
                    wf_run["runner_type"],
                    wf_run["token_permissions"],
                    runner_group=wf_run["runner_group"],
                    machine_name=wf_run["machine_name"],
                    labels=wf_run["requested_labels"],
                    non_ephemeral=wf_run["non_ephemeral"],
                )

                repository.add_accessible_runner(runner)
            runner_detected = True

        return runner_detected

    def enumerate_repository(self, repository: Repository):
        """Enumerate a repository, and check everything relevant to
        self-hosted runner abuse that that the user has permissions to check.

        Args:
            repository (Repository): Wrapper object created from calling the
            API and retrieving a repository.
        """
        if not repository.can_pull():
            Output.error("The user cannot pull, skipping.")
            return
        Output.tabbed(f"Checking repository: {Output.bright(repository.name)}")

        repository.update_time()
        if repository.get_risks():
            for risk in repository.get_risks():
                ActionsReport.report_actions_risk(risk)

        if repository.is_admin():
            runners = self.api.get_repo_runners(repository.name)

            if runners:
                repo_runners = [
                    Runner(
                        runner["name"],
                        machine_name="N/A",
                        os=runner["os"],
                        status=runner["status"],
                        labels=[label["name"] for label in runner["labels"]],
                        runner_type="Repository",
                        runner_group="N/A",
                    )
                    for runner in runners
                ]

                repository.set_runners(repo_runners)
        else:
            runner_wfs = repository.get_sh_workflow_names()
            if runner_wfs:
                Output.info(f"Analyizing run logs for {repository.name}")
                runner_detected = self.perform_runlog_enumeration(
                    repository, runner_wfs
                )
                if runner_detected:
                    RunnersReport.report_runners(repository)

    def enumerate_repository_secrets(self, repository: Repository):
        """Enumerate secrets accessible to a repository.

        Args:
            repository (Repository): Wrapper object created from calling the
            API and retrieving a repository.
        """
        if repository.can_push():
            secrets = self.api.get_secrets(repository.name)
            wrapped_env_secrets = []
            for environment in repository.repo_data["environments"]:
                env_secrets = self.api.get_environment_secrets(
                    repository.name, environment
                )
                for secret in env_secrets:
                    wrapped_env_secrets.append(
                        Secret(secret, repository.name, environment)
                    )

            repo_secrets = [Secret(secret, repository.name) for secret in secrets]

            repo_secrets.extend(wrapped_env_secrets)
            repository.set_secrets(repo_secrets)

            org_secrets = self.api.get_repo_org_secrets(repository.name)
            org_secrets = [
                Secret(secret, repository.org_name) for secret in org_secrets
            ]

            if org_secrets:
                repository.set_accessible_org_secrets(org_secrets)



import pytest
import networkx as nx
from unittest.mock import MagicMock, ANY
from gatox.workflow_graph.graph.tagged_graph import TaggedGraph


class MockNode:
    def __init__(self, name, needs=None, tags=None):
        self.name = name
        self._needs = needs or []
        self._tags = set(tags or [])

    def get_needs(self):
        return self._needs

    def get_tags(self):
        return self._tags

    def __repr__(self):
        return f"MockNode({self.name})"


def test_tagged_graph_init():
    builder = MagicMock()
    graph = TaggedGraph(builder=builder)
    assert graph.builder == builder
    assert isinstance(graph, nx.DiGraph)


def test_add_node():
    builder = MagicMock()
    graph = TaggedGraph(builder=builder)
    node = MockNode("A", tags=["tag1"])
    graph.add_node(node)
    assert node in graph
    assert "tag1" in graph.tags
    assert node in graph.tags["tag1"]


def test_add_tag():
    builder = MagicMock()
    graph = TaggedGraph(builder=builder)
    node = MockNode("A")
    graph.add_node(node)
    graph.add_tag("new_tag", [node])
    assert "new_tag" in graph.tags
    assert node in graph.tags["new_tag"]


def test_remove_tag():
    builder = MagicMock()
    graph = TaggedGraph(builder=builder)
    node = MockNode("A", tags=["tag1"])
    graph.add_node(node)
    graph.remove_tag("tag1")
    assert "tag1" not in graph.tags


def test_remove_node():
    builder = MagicMock()
    graph = TaggedGraph(builder=builder)
    node = MockNode("A", tags=["tag1", "tag2"])
    graph.add_node(node)
    graph.remove_node(node)
    assert node not in graph
    assert node not in graph.get_nodes_by_tag("tag1")
    assert node not in graph.get_nodes_by_tag("tag2")


def test_get_nodes_by_tag():
    builder = MagicMock()
    graph = TaggedGraph(builder=builder)
    node1 = MockNode("A", tags=["tag1"])
    node2 = MockNode("B", tags=["tag2"])
    graph.add_node(node1)
    graph.add_node(node2)
    assert graph.get_nodes_by_tag("tag1") == {node1}
    assert graph.get_nodes_by_tag("tag2") == {node2}


def test_add_tags_to_node():
    builder = MagicMock()
    graph = TaggedGraph(builder=builder)
    node = MockNode("A", tags=["initial"])
    graph.add_node(node)
    graph.add_tags_to_node(node, ["extra"])
    assert "extra" in graph.tags
    assert node in graph.tags["extra"]


def test_remove_tags_from_node():
    builder = MagicMock()
    graph = TaggedGraph(builder=builder)
    node = MockNode("A", tags=["tag1", "tag2"])
    graph.add_node(node)
    graph.remove_tags_from_node(node, ["tag1"])
    assert "tag1" not in graph.get_tags_for_node(node)
    assert "tag2" in graph.get_tags_for_node(node)


def test_dfs_to_tag():
    builder = MagicMock()
    builder.initialize_node = MagicMock()
    graph = TaggedGraph(builder=builder)

    nodeA = MockNode("A", needs=[], tags=["start"])
    nodeB = MockNode("B", needs=[nodeA], tags=["middle"])
    nodeC = MockNode("C", needs=[nodeB], tags=["target"])
    graph.add_node(nodeA)
    graph.add_node(nodeB)
    graph.add_node(nodeC)
    graph.add_edge(nodeA, nodeB)
    graph.add_edge(nodeB, nodeC)

    paths = graph.dfs_to_tag(nodeA, "target", api=MagicMock())
    assert len(paths) == 1
    assert paths[0] == [nodeA, nodeB, nodeC]



"""
Copyright 2025, Adnan Khan

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import json
from typing import IO

from gatox.models.workflow import Workflow
from gatox.models.repository import Repository


class CacheManager:
    """
    Singleton class that manages an in-memory cache for workflows and reusable actions.

    """

    _instance = None

    def __new__(cls):
        """
        Create a new instance of the class. If an instance already exists, return that instance.
        """
        if cls._instance is None:
            cls._instance = super(CacheManager, cls).__new__(cls)
            cls._instance.repo_wf_lookup = {}
            cls._instance.repo_store = {}
            cls._instance.workflow_cache = {}
            cls._instance.action_cache = {}
        return cls._instance

    def get_repos(self) -> list:
        """
        Get all repository names from the in-memory dictionary.

        Returns:
            list: A list of repository names.
        """
        return list(self.repo_store.keys())

    def get_workflow(self, repo_slug: str, workflow_name: str) -> Workflow:
        """
        Get a workflow from the in-memory dictionary.

        Args:
            repo_slug (str): The repository slug.
            workflow_name (str): The name of the workflow.

        Returns:
            Workflow: The workflow object if found, else None.
        """
        key = f"{repo_slug.lower()}:{workflow_name}"
        return self.workflow_cache.get(key, None)

    def is_repo_cached(self, repo_slug: str):
        """
        Check if a repository is in the in-memory dictionary.

        Args:
            repo_slug (str): The repository slug.

        Returns:
            bool: True if the repository is cached, else False.
        """
        return repo_slug.lower() in self.repo_wf_lookup

    def is_action_cached(self, repo_slug: str, action_path: str, ref: str):
        """
        Check if action is cached.

        Args:
            repo_slug (str): The repository slug.
            action_path (str): The path to the action.
            ref (str): The reference (e.g., branch or tag).

        Returns:
            bool: True if the action is cached, else False.
        """
        key = f"{repo_slug.lower()}:{action_path}:{ref}"
        return key in self.action_cache

    def get_workflows(self, repo_slug: str) -> list:
        """
        Get all workflows for a repository from the in-memory dictionary.

        Args:
            repo_slug (str): The repository slug.

        Returns:
            list: A list of workflow objects.
        """
        repo_slug = repo_slug.lower()
        wf_keys = self.repo_wf_lookup.get(repo_slug, None)
        if wf_keys:
            return [
                self.workflow_cache[f"{repo_slug.lower()}:{key}"] for key in wf_keys
            ]
        else:
            return set()

    def get_action(self, repo_slug: str, action_path: str, ref: str):
        """
        Get an action from the in-memory dictionary.

        Args:
            repo_slug (str): The repository slug.
            action_path (str): The path to the action.
            ref (str): The reference (e.g., branch or tag).

        Returns:
            str: The action contents if found, else None.
        """
        key = f"{repo_slug.lower()}:{action_path}:{ref}"
        return self.action_cache.get(key, None)

    def set_repository(self, repository: Repository):
        """
        Set a repository in the in-memory dictionary.

        Args:
            repository (Repository): The repository object to cache.
        """
        key = repository.name.lower()
        self.repo_store[key] = repository

    def get_repository(self, repo_slug: str) -> Repository:
        """
        Get a repository from the in-memory dictionary.

        Args:
            repo_slug (str): The repository slug.

        Returns:
            Repository: The repository object if found, else None.
        """
        return self.repo_store.get(repo_slug.lower(), None)

    def set_workflow(self, repo_slug: str, workflow_name: str, value: Workflow):
        """
        Set a workflow in the in-memory dictionary.

        Args:
            repo_slug (str): The repository slug.
            workflow_name (str): The name of the workflow.
            value (Workflow): The workflow object to cache.
        """
        key = f"{repo_slug.lower()}:{workflow_name}"
        if repo_slug.lower() not in self.repo_wf_lookup:
            self.repo_wf_lookup[repo_slug.lower()] = set()
        self.repo_wf_lookup[repo_slug.lower()].add(workflow_name)
        self.workflow_cache[key] = value

    def set_empty(self, repo_slug: str):
        """
        Set an empty value in the in-memory dictionary for a repository.

        Args:
            repo_slug (str): The repository slug.
        """
        self.repo_wf_lookup[repo_slug.lower()] = set()

    def set_action(self, repo_slug: str, action_path: str, ref: str, value: str):
        """
        Set an action in the in-memory dictionary.

        Args:
            repo_slug (str): The repository slug.
            action_path (str): The path to the action.
            ref (str): The reference (e.g., branch or tag).
            value (str): The action contents to cache.
        """
        key = f"{repo_slug.lower()}:{action_path}:{ref}"
        self.action_cache[key] = value

    def serialize_full_cache(self, output_stream: IO[str]) -> None:
        """
        Serializes the action cache to JSON and writes it to the provided output stream.
        Raises ValueError if the stream is invalid.
        """
        if not hasattr(output_stream, "write"):
            raise ValueError("Output stream must have a 'write' method.")

        json.dump(self._instance.action_cache, output_stream, indent=2)
        output_stream.flush()



import pytest
from gatox.workflow_graph.nodes.job import JobNode


def test_job_node_init():
    """Test JobNode initialization"""
    job = JobNode("test_job", "main", "org/repo", ".github/workflows/test.yml")

    assert job.name == "org/repo:main:.github/workflows/test.yml:test_job"
    assert job.ref == "main"
    assert job.repo_name() == "org/repo"
    assert job.params == {}
    assert job.if_condition is None
    assert job.wf_reference is None
    assert job.needs == []
    assert job.deployments == []
    assert job.self_hosted is False
    assert job.outputs == {}


def test_job_node_hash():
    """Test JobNode hash"""
    job = JobNode("test_job", "main", "org/repo", ".github/workflows/test.yml")
    assert hash(job) == hash((job.name, "JobNode"))


def test_check_selfhosted():
    """Test self-hosted runner detection"""
    job = JobNode("test_job", "main", "org/repo", ".github/workflows/test.yml")

    # Test direct self-hosted label
    assert job._check_selfhosted({"runs-on": "self-hosted"}) is True

    # Test non self-hosted runner
    assert job._check_selfhosted({"runs-on": "ubuntu-latest"}) is False


def test_get_workflow_name():
    """Test getting workflow name"""
    job = JobNode("test_job", "main", "org/repo", ".github/workflows/test.yml")
    assert job.get_workflow_name() == "test.yml"


def test_populate():
    """Test populating job node with definition"""
    job = JobNode("test_job", "main", "org/repo", ".github/workflows/test.yml")

    job_def = {
        "runs-on": "self-hosted",
        "with": {"param1": "value1"},
        "environment": "prod",
        "env": {"ENV_VAR": "value"},
        "outputs": {"output1": "value1"},
    }

    job.populate(job_def, None)

    assert job.self_hosted is True
    assert job.params == {"param1": "value1"}
    assert job.deployments == ["prod"]
    assert job.get_env_vars() == {"ENV_VAR": "value"}
    assert job.outputs == {"output1": "value1"}


def test_job_node_equality():
    """Test JobNode equality comparison"""
    job1 = JobNode("test_job", "main", "org/repo", ".github/workflows/test.yml")
    job2 = JobNode("test_job", "main", "org/repo", ".github/workflows/test.yml")
    job3 = JobNode("other_job", "main", "org/repo", ".github/workflows/test.yml")

    assert job1 == job2
    assert job1 != job3


def test_get_needs():
    """Test getting job needs"""
    job = JobNode("test_job", "main", "org/repo", ".github/workflows/test.yml")

    need = JobNode("need_job", "main", "org/repo", ".github/workflows/test.yml")
    job.add_needs(need)

    assert job.get_needs() == [need]


def test_get_tags():
    """Test getting job tags"""
    job = JobNode("test_job", "main", "org/repo", ".github/workflows/test.yml")

    # Default tags
    assert job.get_tags() == {"JobNode"}

    # With self-hosted runner
    job.self_hosted = True
    assert job.get_tags() == {"JobNode", "self-hosted"}


def test_get_attrs():
    """Test getting job attributes"""
    job = JobNode("test_job", "main", "org/repo", ".github/workflows/test.yml")

    expected = {"JobNode": True, "is_soft_gate": False, "is_hard_gate": False}

    assert job.get_attrs() == expected



from gatox.cli.colors import Fore, Style
from gatox.util.arg_utils import StringType


def configure_parser_search(parser):
    """Helper method to add arguments to the search subparser.

    Args:
        parser: Add arguments to the search module subparser.
    """
    parser.add_argument(
        "--target",
        "-t",
        help="Organization to enumerate using GitHub code search.",
        metavar=f"{Fore.RED}ORGANIZATION{Style.RESET_ALL}",
        required=False,
    )

    parser.add_argument(
        "--query",
        "-q",
        help="Pass a custom query to GitHub code search",
        metavar="QUERY",
        required=False,
    )

    parser.add_argument(
        "--sourcegraph",
        "-sg",
        help="Use Sourcegraph API to search for self-hosted runners.",
        required=False,
        action="store_true",
    )

    parser.add_argument(
        "--output-text",
        "-oT",
        help=("Save enumeration output to text file."),
        metavar="TEXT_FILE",
        type=StringType(256),
    )



import re

from unittest.mock import patch
from unittest.mock import MagicMock
from gatox.attack.attack import Attacker
from gatox.cli.output import Output

output = Output(True)


# From https://stackoverflow.com/questions/14693701/
# how-can-i-remove-the-ansi-escape-sequences-from-a-string-in-python
def escape_ansi(line):
    ansi_escape = re.compile(r"(?:\x1B[@-_]|[\x80-\x9F])[0-?]*[ -/]*[@-~]")
    return ansi_escape.sub("", line)


def test_init():
    """Test constructor for enumerator."""

    gh_attacker = Attacker(
        "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        socks_proxy=None,
        http_proxy="localhost:8080",
    )

    assert gh_attacker.http_proxy == "localhost:8080"


# @patch("gatox.attack.attack.time.sleep")
# @patch("gatox.attack.attack.Api")
# @patch("gatox.attack.attack.Git")
# def test_fork_pr(mock_git, mock_api, mock_time, capsys):
#     """Test creating a malicious fork PR
#     """

#     mock_api.return_value.check_user.return_value = {
#         "user": 'testUser',
#         "name": 'test user',
#         "scopes": ['repo','workflow']
#     }

#     mock_api.return_value.get_recent_workflow.return_value = \
#         12345

#     mock_api.return_value.fork_repository.return_value = \
#         'testOrg/targetRepo'
#     mock_api.return_value.create_fork_pr.return_value = \
#         'https://github.com/testOrg/targetRepo/pulls/12'

#     mock_api.return_value.proxies = {
#         "https": "http://localhost:8080"
#     }

#     gh_attacker = Attacker(

#         "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
#         socks_proxy=None,
#         http_proxy="localhost:8080"
#     )

#     gh_attacker.fork_pr_attack('targetRepo', 'develop', 'Bad PR', 'attack',
#                                'whoami', None, 'message')

#     captured = capsys.readouterr()

#     print_output = captured.out

#     assert "Successfully created fork: testOrg/targetRepo" in escape_ansi(
#         print_output
#     )

#     assert " viewed at: https://github.com/testOrg/targetRepo/pulls/12" in \
#         escape_ansi(print_output)

# @patch("gatox.attack.attack.time.sleep")
# @patch("gatox.attack.attack.Api")
# @patch("gatox.attack.attack.Git")
# def test_fork_pr_timeout(mock_git, mock_api, mock_time, capsys):
#     """Test creating a malicious fork PR
#     """

#     mock_api.return_value.check_user.return_value = {
#         "user": 'testUser',
#         "name": 'test user',
#         "scopes": ['repo','workflow']
#     }

#     mock_api.return_value.fork_repository.return_value = \
#         'testOrg/targetRepo'

#     mock_api.return_value.get_repository.return_value = \
#         None

#     mock_api.return_value.proxies = {
#         "https": "http://localhost:8080"
#     }

#     gh_attacker = Attacker(

#         "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
#         socks_proxy=None,
#         http_proxy="localhost:8080"
#     )

#     status = gh_attacker.fork_pr_attack(
#         'targetRepo', 'develop', 'Bad PR',
#         'attack', 'whoami', None, 'message'
#     )

#     captured = capsys.readouterr()

#     print_output = captured.out

#     assert status is False
#     assert "Forked repository not found after 30 seconds!" in escape_ansi(
#         print_output
#     )

# @patch("gatox.attack.attack.Api")
# @patch("gatox.attack.attack.Git")
# def test_fork_pr_perm(mock_git, mock_api, capsys):
#     """Test creating a malicious fork PR
#     """

#     mock_api.return_value.check_user.return_value = {
#         "user": 'testUser',
#         "name": 'test user',
#         "scopes": ['repo']
#     }

#     mock_api.return_value.proxies = {
#         "https": "http://localhost:8080"
#     }

#     gh_attacker = Attacker(

#         "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
#         socks_proxy=None,
#         http_proxy="localhost:8080"
#     )

#     gh_attacker.fork_pr_attack('targetRepo', 'develop', 'Bad PR', 'attack',
#                                'whoami', None, 'message')

#     captured = capsys.readouterr()

#     print_output = captured.out

#     assert " The user does not have the necessary scopes" in \
#         escape_ansi(print_output)


@patch("gatox.attack.attack.time.sleep")
@patch("gatox.attack.attack.Api")
def test_push_workflow_attack(mock_api, mock_time, capsys):
    """Test the shell workflow attack."""

    mock_api.return_value.check_user.return_value = {
        "user": "testUser",
        "name": "test user",
        "scopes": ["repo", "workflow"],
    }

    mock_api.return_value.proxies = {"https": "http://localhost:8080"}

    mock_api.return_value.create_branch.return_value = True
    mock_api.return_value.commit_file.return_value = (
        "8933f8abb60e4e02ae1b8dd3f109bc7b6812e54f"
    )
    mock_api.return_value.get_recent_workflow.return_value = 1
    mock_api.return_value.get_workflow_status.return_value = 1
    mock_api.return_value.delete_branch.return_value = True

    gh_attacker = Attacker(
        "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        socks_proxy=None,
        http_proxy="localhost:8080",
    )

    gh_attacker.push_workflow_attack(
        "targetRepo", "whoami", None, None, "message", True
    )

    captured = capsys.readouterr()

    print_output = captured.out

    assert "Workflow logs downloaded to" in escape_ansi(print_output)
    assert "Workflow still incomplete but hit timeout!" not in escape_ansi(print_output)


@patch("gatox.attack.attack.Api")
def test_push_workflow_attack_perm(mock_api, capsys):
    """Test executing shell workflow attack with invalid permissions."""

    mock_api.return_value.check_user.return_value = {
        "user": "testUser",
        "name": "test user",
        "scopes": ["repo"],
    }

    mock_api.return_value.proxies = {"https": "http://localhost:8080"}

    gh_attacker = Attacker(
        "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        socks_proxy=None,
        http_proxy="localhost:8080",
    )

    gh_attacker.push_workflow_attack(
        "targetRepo", "whoami", None, None, "message", False
    )

    captured = capsys.readouterr()

    print_output = captured.out

    assert " The user does not have the necessary scopes" in escape_ansi(print_output)


@patch("gatox.attack.attack.time.sleep")
@patch("gatox.attack.attack.Api")
def test_push_workflow_attack_fail_wf(mock_api, mock_time, capsys):
    """Test the shell workflow attack."""

    mock_api.return_value.check_user.return_value = {
        "user": "testUser",
        "name": "test user",
        "scopes": ["repo", "workflow"],
    }

    mock_api.return_value.proxies = {"https": "http://localhost:8080"}

    mock_api.return_value.create_branch.return_value = True
    mock_api.return_value.commit_file.return_value = (
        "8933f8abb60e4e02ae1b8dd3f109bc7b6812e54f"
    )
    mock_api.return_value.get_recent_workflow.return_value = -1
    mock_api.return_value.get_workflow_status.return_value = 0
    mock_api.return_value.delete_branch.return_value = True

    gh_attacker = Attacker(
        "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        socks_proxy=None,
        http_proxy="localhost:8080",
    )

    gh_attacker.push_workflow_attack(
        "targetRepo", "whoami", None, None, "message", True
    )

    captured = capsys.readouterr()

    print_output = captured.out

    assert "Failed to find the created workflow!" in escape_ansi(print_output)


@patch("gatox.attack.attack.time.sleep")
@patch("gatox.attack.attack.Api")
def test_push_workflow_attack_fail_timeout(mock_api, mock_time, capsys):
    """Test the shell workflow attack."""

    mock_api.return_value.check_user.return_value = {
        "user": "testUser",
        "name": "test user",
        "scopes": ["repo", "workflow"],
    }

    mock_api.return_value.proxies = {"https": "http://localhost:8080"}

    mock_api.return_value.create_branch.return_value = True
    mock_api.return_value.commit_file.return_value = (
        "8933f8abb60e4e02ae1b8dd3f109bc7b6812e54f"
    )
    mock_api.return_value.get_recent_workflow.return_value = 0
    mock_api.return_value.get_workflow_status.return_value = 0
    mock_api.return_value.delete_branch.return_value = True

    gh_attacker = Attacker(
        "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        socks_proxy=None,
        http_proxy="localhost:8080",
    )

    gh_attacker.push_workflow_attack(
        "targetRepo", "whoami", None, None, "message", True
    )

    captured = capsys.readouterr()

    print_output = captured.out

    assert "Failed to find the created workflow!" in escape_ansi(print_output)


@patch("gatox.attack.attack.time.sleep")
@patch("gatox.attack.attack.Api")
def test_push_workflow_attack_fail_timeout2(mock_api, mock_time, capsys):
    """Test the shell workflow attack."""

    mock_api.return_value.check_user.return_value = {
        "user": "testUser",
        "name": "test user",
        "scopes": ["repo", "workflow"],
    }

    mock_api.return_value.proxies = {"https": "http://localhost:8080"}

    mock_api.return_value.create_branch.return_value = True
    mock_api.return_value.commit_file.return_value = (
        "8933f8abb60e4e02ae1b8dd3f109bc7b6812e54f"
    )
    mock_api.return_value.get_recent_workflow.return_value = 1
    mock_api.return_value.get_workflow_status.return_value = 0
    mock_api.return_value.delete_branch.return_value = True

    gh_attacker = Attacker(
        "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        socks_proxy=None,
        http_proxy="localhost:8080",
    )

    gh_attacker.push_workflow_attack(
        "targetRepo", "whoami", None, None, "message", True
    )

    captured = capsys.readouterr()

    print_output = captured.out

    assert "The workflow is incomplete but hit the timeout" in escape_ansi(print_output)


@patch("gatox.attack.attack.time.sleep")
@patch("gatox.attack.attack.Api")
def test_push_workflow_attack_fail_branch(mock_api, mock_time, capsys):
    """Test the shell workflow attack."""

    mock_api.return_value.check_user.return_value = {
        "user": "testUser",
        "name": "test user",
        "scopes": ["repo", "workflow"],
    }

    mock_api.return_value.proxies = {"https": "http://localhost:8080"}

    mock_api.return_value.get_repo_branch.return_value = -1

    gh_attacker = Attacker(
        "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        socks_proxy=None,
        http_proxy="localhost:8080",
    )

    gh_attacker.push_workflow_attack(
        "targetRepo", "whoami", None, None, "message", True
    )

    captured = capsys.readouterr()

    print_output = captured.out

    assert "Failed to check for remote branch!" in escape_ansi(print_output)


@patch("gatox.attack.attack.time.sleep")
@patch("gatox.attack.attack.Api")
def test_push_workflow_attack_fail_branch2(mock_api, mock_time, capsys):
    """Test the shell workflow attack."""

    mock_api.return_value.check_user.return_value = {
        "user": "testUser",
        "name": "test user",
        "scopes": ["repo", "workflow"],
    }

    mock_api.return_value.proxies = {"https": "http://localhost:8080"}

    mock_api.return_value.get_repo_branch.return_value = 1

    gh_attacker = Attacker(
        "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        socks_proxy=None,
        http_proxy="localhost:8080",
    )

    gh_attacker.push_workflow_attack(
        "targetRepo", "whoami", None, None, "message", True
    )

    captured = capsys.readouterr()

    print_output = captured.out

    assert "Remote branch, " in escape_ansi(print_output)
    assert ", already exists!" in escape_ansi(print_output)



import base64
import copy
import time
import requests
import logging
import zipfile
import re
import io

from gatox.cli.output import Output
from datetime import datetime, timezone, timedelta
from gatox.enumerate.ingest.ingest import DataIngestor
from gatox.models.workflow import Workflow
from gatox.github.gql_queries import GqlQueries

logger = logging.getLogger(__name__)


class Api:
    """Class to serve as an abstraction layer to interact with the GitHub API.
    It handles utilizing proxies, along with passing the PAT and handling any
    rate limiting or network issues.
    """

    RUNNER_RE = re.compile(r"Runner name: \'([\w+-.]+)\'")
    MACHINE_RE = re.compile(r"Machine name: \'([\w+-.]+)\'")
    RUNNERGROUP_RE = re.compile(r"Runner group name: \'([\w+-.]+)\'")
    RUNNERTYPE_RE = re.compile(r"([\w+-.]+)")

    RUN_THRESHOLD = 90

    def __init__(
        self,
        pat: str,
        version: str = "2022-11-28",
        http_proxy: str = None,
        socks_proxy: str = None,
        github_url: str = "https://api.github.com",
    ):
        """Initialize the API abstraction layer to interact with the GitHub
        REST API.

        Args:
            pat (str): GitHub personal access token that will be used for API
            calls.
            version (str): API version to use that will be passed with the
            X-GitHub-Api-Version header.
            http_proxy (str, optional): HTTP Proxy to use for API calls.
            Defaults to None.
            socks_proxy (str, optional): SOCKS Proxy to use for API calls.
            Defaults to None.
        """
        self.pat = pat
        self.proxies = None
        self.verify_ssl = True
        self.headers = {
            "Accept": "application/vnd.github+json",
            "Authorization": f"Bearer {pat}",
            "X-GitHub-Api-Version": version,
        }
        if not github_url:
            self.github_url = "https://api.github.com"
        else:
            self.github_url = github_url

        if http_proxy and socks_proxy:
            raise ValueError(
                "A SOCKS & HTTP proxy cannot be used at the same "
                "time! Please pass only one!"
            )

        if http_proxy:
            # We are likely using BURP, so disable SSL.
            requests.packages.urllib3.disable_warnings()
            self.verify_ssl = False
            self.proxies = {
                "http": f"http://{http_proxy}",
                "https": f"http://{http_proxy}",
            }
        elif socks_proxy:
            self.proxies = {
                "http": f"socks5://{socks_proxy}",
                "https": f"socks5://{socks_proxy}",
            }

        if self.github_url != "https://api.github.com":
            self.verify_ssl = False
            requests.packages.urllib3.disable_warnings()

    def __check_rate_limit(self, headers):
        """Checks the rate limit, and pauses Gato execution until the rate
        limit resets.
        """
        if (
            "X-Ratelimit-Remaining" in headers
            and int(headers["X-Ratelimit-Remaining"])
            < int(headers["X-RateLimit-Limit"]) // 20
            and headers["X-Ratelimit-Resource"] == "core"
        ):
            gh_date = headers["Date"]
            reset_utc = int(headers["X-Ratelimit-Reset"])
            # Convert date to UTC
            date = datetime.strptime(gh_date, "%a, %d %b %Y %H:%M:%S %Z")
            date = date.replace(tzinfo=timezone.utc)
            reset_time = date.fromtimestamp(reset_utc, tz=timezone.utc)

            sleep_time = (reset_time - date).seconds
            sleep_time_mins = str(sleep_time // 60)

            # Yes, we are breaking the "don't print in API code" rule; however,
            # the alternative would be to handle a rate limit exception in
            # all calling code. We inform the here user that we are sleeping.
            # very large orgs will take several hours to enumerate, especially
            # if runlog enumeration is enabled.
            Output.warn(
                f"Sleeping for {Output.bright( sleep_time_mins + ' minutes')} "
                "to prevent rate limit exhaustion!"
            )

            time.sleep(sleep_time + 1)

    def __process_run_log(self, log_content: bytes, run_info: dict):
        """Utility method to process a run log zip file.

        Args:
            log_content (bytes): Zip file downloaded from GitHub
            run_info (dict): Metadata about the run from the GitHub API
        Returns:
            dict: metadata about the run execution.
        """
        log_package = dict()
        token_permissions = dict()
        runner_type = None
        non_ephemeral = False
        labels = []
        runner_name = None
        machine_name = None
        runner_group = None

        with zipfile.ZipFile(io.BytesIO(log_content)) as runres:
            for zipinfo in runres.infolist():
                if re.match("[0-9]{1}_.*", zipinfo.filename):
                    with runres.open(zipinfo) as run_setup:
                        content = run_setup.read().decode()
                        content_lines = content.split("\n")
                        if (
                            "Image Release: https://github.com/actions/runner-images"
                            in content
                            or "Job is about to start running on the hosted runner: GitHub Actions"
                            in content
                        ) and not "1ES.Pool" in content:
                            # Larger runners will appear to be self-hosted, but
                            # they will have the image name. Skip if we see this.
                            # If the log contains "job is about to start running on hosted runner",
                            # the runner is a Github hosted runner so we can skip it.
                            continue
                        elif (
                            "Self-hosted runners in the repository are disabled"
                            in content
                        ):
                            break
                        index = 0
                        while index < len(content_lines) and content_lines[index]:
                            line = content_lines[index]

                            if "Requested labels: " in line:
                                labels = line.split("Requested labels: ")[1].split(", ")

                            if "Runner name: " in line:
                                runner_name = line.split("Runner name: ")[1].replace(
                                    "'", ""
                                )

                            if "Machine name: " in line:
                                machine_name = line.split("Machine name: ")[1].replace(
                                    "'", ""
                                )

                            if "Runner group name:" in line:
                                runner_group = line.split("Runner group name: ")[
                                    1
                                ].replace("'", "")

                            if "Job is about to start running on" in line:
                                runner_type = line.split()[-1]
                                matches = Api.RUNNERTYPE_RE.search(runner_type)
                                runner_type = matches.group(1)

                            if "GITHUB_TOKEN Permission" in line:
                                while "[endgroup]" not in content_lines[index + 1]:
                                    index += 1
                                    scope = (
                                        content_lines[index].split()[1].replace(":", "")
                                    )
                                    permission = content_lines[index].split()[2]
                                    token_permissions[scope] = permission
                                log_package["token_permissions"] = token_permissions

                            if "Cleaning the repository" in line:
                                non_ephemeral = True
                            log_package["non_ephemeral"] = non_ephemeral

                            index += 1

                        # Continue if there is no runner name. This means
                        # we picked up a pending workflow.
                        if not runner_name:
                            continue

                        log_package = {
                            "requested_labels": labels,
                            "runner_name": runner_name,
                            "machine_name": machine_name,
                            "runner_group": runner_group,
                            "runner_type": runner_type,
                            "run_id": run_info["id"],
                            "run_attempt": run_info["run_attempt"],
                            "non_ephemeral": non_ephemeral,
                            "token_permissions": token_permissions,
                        }

                    return log_package

    def __get_full_runlog(self, log_content: bytes, run_name: str):
        """Gets the full text of the runlog from the zip file by matching the
         filename.

        Args:
            log_content (bytes): zip file binary
            run_name (str): Name of file to look for

        Returns:
            str: Runlog content
        """
        with zipfile.ZipFile(io.BytesIO(log_content)) as runres:
            for zipinfo in runres.infolist():
                if f"0_{run_name}" in zipinfo.filename:
                    with runres.open(zipinfo) as run_log:
                        content = run_log.read().decode()

                        return content

    def __get_raw_file(self, repo: str, file_path: str, ref: str):
        """Get a raw file with a web request."""
        resp = requests.get(
            f"https://raw.githubusercontent.com/{repo}/{ref}/{file_path}",
            proxies=self.proxies,
            verify=self.verify_ssl,
        )

        if resp.status_code == 404:
            return None
        elif resp.status_code == 200:
            return resp.text

    @staticmethod
    def __verify_result(response: requests.Response, expected_code: int):
        """Verifies that the response matches the expected code. If it does not
        match, then the response is logged and the program exits.

        Args:
            response (requests.Response): Response object from a request.
            expected_code (int): Expected status code from the request.
        """
        if response.status_code != expected_code:
            logger.warning(
                f"Expected status code {expected_code}, but got "
                f"{response.status_code}!"
            )
            logger.debug(response.text)
            return False
        return True

    def is_app_token(self):
        """Returns if the API is using a GitHub App installation token."""
        return self.pat.startswith("ghs_")

    def call_get(self, url: str, params: dict = None, strip_auth=False):
        """Internal method to wrap a GET request so that proxies and headers
        do not need to be repeated.

        Args:
            url (str): Url path for the API request
            params (dict, optional): Parameters to pass to the request.
            strip_auth (bool): Whether to make the request without any auth
            token. Defaults to False.
            Defaults to None.

        Returns:
            Response: Returns the requests response object.
        """
        request_url = self.github_url + url

        get_header = copy.deepcopy(self.headers)
        if strip_auth:
            del get_header["Authorization"]

        for i in range(0, 5):
            try:
                logger.debug(f"Making GET API request to {request_url}!")
                api_response = requests.get(
                    request_url,
                    headers=get_header,
                    proxies=self.proxies,
                    params=params,
                    verify=self.verify_ssl,
                )
                break
            except Exception:
                logger.warning("GET request failed due to transport error re-trying!")
                continue

        self.__check_rate_limit(api_response.headers)

        return api_response

    def call_post(self, url: str, params: dict = None):
        """Internal method to wrap a POST request so that proxies and headers
        do not need to be updated in each method.

        Args:
            url (str): URL path to make POST request to.
            params (dict, optional): Parameters to send as part of the request.
            Defaults to None.
        Returns:
            Response: Returns the requests response object.
        """
        request_url = self.github_url + url
        logger.debug(f"Making POST API request to {request_url}!")

        api_response = requests.post(
            request_url,
            headers=self.headers,
            proxies=self.proxies,
            json=params,
            verify=self.verify_ssl,
        )
        logger.debug(
            f"The POST request to {request_url} returned a "
            f"{api_response.status_code}!"
        )

        self.__check_rate_limit(api_response.headers)

        return api_response

    def call_patch(self, url: str, params: dict = None):
        """Internal method to wrap a PATCH request so that proxies and headers
        do not need to be updated in each method.

        Args:
            url (str): URL path to make PATCH request to.
            params (dict, optional): Parameters to send as part of the request.
            Defaults to None.
        Returns:
            Response: Returns the requests response object.
        """
        request_url = self.github_url + url
        logger.debug(f"Making PATCH API request to {request_url}!")

        api_response = requests.patch(
            request_url,
            headers=self.headers,
            proxies=self.proxies,
            json=params,
            verify=self.verify_ssl,
        )
        logger.debug(
            f"The PATCH request to {request_url} returned a "
            f"{api_response.status_code}!"
        )

        self.__check_rate_limit(api_response.headers)

        return api_response

    def call_put(self, url: str, params: dict = None):
        """Internal method to wrap a PUT request so that proxies and headers
        do not need to be updated in each method.

        Args:
            url (stre): _description_
            params (dict, optional): _description_. Defaults to None.
        """
        request_url = self.github_url + url
        logger.debug(f"Making PUT API request to {request_url}!")

        api_response = requests.put(
            request_url,
            headers=self.headers,
            proxies=self.proxies,
            json=params,
            verify=self.verify_ssl,
        )

        self.__check_rate_limit(api_response.headers)

        return api_response

    def call_delete(self, url: str, params: dict = None):
        """Internal method to wrap a POST request so that proxies and headers
        do not need to be updated in each method.

        Args:
            url (str): URL path to make POST request to.
            params (dict, optional): Parameters to send as part of the request.
            Defaults to None.
        Returns:
            Response: Returns the requests response object.
        """
        request_url = self.github_url + url
        logger.debug(f"Making DELETE API request to {request_url}!")

        api_response = requests.delete(
            request_url,
            headers=self.headers,
            proxies=self.proxies,
            json=params,
            verify=self.verify_ssl,
        )
        logger.debug(
            f"The POST request to {request_url} returned a "
            f"{api_response.status_code}!"
        )

        self.__check_rate_limit(api_response.headers)

        return api_response

    def delete_repository(self, repo_name: str):
        """Deletes the provided repository, if the user has administrative
        permissions on that repository.

        Args:
            repo_name (str): Name of repository to delete in Org/Owner format.
        Returns:
            bool: True if the repository was deleted, False otherwise.
        """
        result = self.call_delete(f"/repos/{repo_name}")

        if result.status_code == 204:
            logger.info(f"Successfully deleted {repo_name}!")
        else:
            logger.warning(f"Unable to delete repository {repo_name}!")
            return False

        return True

    def fork_repository(self, repo_name: str):
        """Creates a fork of a public repository and returns the name of
        the newly created fork.

        Args:
            repo_name (str): Name of the repository to fork.

        Returns:
            str: Full name of the newly forked repo in User/Repo format. False
            if there was a faiure.
        """
        post_params = {"default_branch_only": True}

        result = self.call_post(f"/repos/{repo_name}/forks", params=post_params)

        if result.status_code == 202:
            fork_info = result.json()
            return fork_info["full_name"]
        elif result.status_code == 403:
            # likely permission error, log it.
            logger.warning("Forking this repository is forbidden!")
            return False
        elif result.status_code == 404:
            logger.warning("Unable to fork due to 404, ensure repository exists.")
            return False
        else:
            logger.warning("Repository fork failed!")
            return False

    def create_fork_pr(
        self,
        target_repo: str,
        source_user: str,
        source_branch: str,
        target_branch: str,
        pr_title: str,
    ):
        """Creates a pull request from source_repo to target_repo. This is

        Args:
            target_repo (str): Target repo  (the one we are targeting)
            source_repo (str): Source repo for the PR (the one we own)
        Returns:
            str: URL of the newly created pull-request.
        """
        pr_params = {
            "title": pr_title,
            "head": f"{source_user}:{source_branch}",
            "base": f"{target_branch}",
            "body": "This is a test pull request created for CI/CD"
            " vulnerability testing purposes.",
            "maintainer_can_modify": False,
            "draft": True,
        }

        result = self.call_post(f"/repos/{target_repo}/pulls", params=pr_params)

        if result.status_code == 201:
            details = result.json()
            return details["html_url"]
        else:
            logger.warning(
                f"Failed to create PR for fork,"
                f" the status code was: {result.status_code}!"
            )
            return None

    def check_organizations(self):
        """Check organizations that the authenticated user belongs to.

        Returns:
            list(str): List of strings containing the organization names that
            the user is a member of.
        """
        organizations = []
        page = 1
        per_page = 100

        while True:
            params = {"page": page, "per_page": per_page}
            result = self.call_get("/user/orgs", params=params)

            if result.status_code == 200:
                orgs = result.json()
                if not orgs:
                    break

                organizations.extend([org["login"] for org in orgs])
                page += 1
            elif result.status_code == 403:
                break
            else:
                break

        return organizations

    def get_repository(self, repository: str):
        """Retrieve a repository using the GitHub API.

        Args:
            repository (str): Repository name in org/Repo format.
        Returns:
            dict: Dictionary containing repository info from the GitHub API.
        """
        result = self.call_get(f"/repos/{repository}")

        if result.status_code == 200:
            return result.json()

    def get_user_type(self, username: str):
        """
        Retrieve the type of a user from the API.

        This function sends a GET request to the API to retrieve information about a user specified by the username.
        If the request is successful (status code 200), it returns the type of the user.

        Args:
            username (str): The username of the user whose type is to be retrieved.

        Returns:
            str: The type of the user if the request is successful.

        Raises:
            requests.exceptions.RequestException: If the request fails due to network issues or invalid responses.
            KeyError: If the 'type' key is not present in the response JSON.
        """
        result = self.call_get(f"/users/{username}")

        if result.status_code == 200:
            return result.json()["type"]

    def get_own_repos(self):
        """Retrieve all repositories where the user is the owner or a collaborator."""

        repos = []

        get_params = {"affiliation": "collaborator,owner", "per_page": 100, "page": 1}

        result = self.call_get("/user/repos", params=get_params)
        if result.status_code == 200:
            listing = result.json()
            repos.extend(
                [repo["full_name"] for repo in listing if not repo["archived"]]
            )

            # Check if there are more pages
            while len(listing) == 100:
                get_params["page"] += 1
                result = self.call_get("/user/repos", params=get_params)
                if result.status_code == 200:
                    listing = result.json()
                    repos.extend(
                        [repo["full_name"] for repo in listing if not repo["archived"]]
                    )
        return repos

    def get_user_repos(self, username: str):
        """Retrieve all repositories belonging to the user."""

        repos = []

        get_params = {"type": "owner", "per_page": 100, "page": 1}

        result = self.call_get(f"/users/{username}/repos", params=get_params)
        if result.status_code == 200:
            listing = result.json()
            repos.extend(
                [repo["full_name"] for repo in listing if not repo["archived"]]
            )

            # Check if there are more pages
            while len(listing) == 100:
                get_params["page"] += 1
                result = self.call_get(f"/users/{username}/repos", params=get_params)
                if result.status_code == 200:
                    listing = result.json()
                    repos.extend(
                        [repo["full_name"] for repo in listing if not repo["archived"]]
                    )
        return repos

    def get_organization_details(self, org: str):
        """Query the GitHub API for details about the specific organization.

        If the token has an org admin scope, then this will reveal additional
        information about the org.

        Args:
            org (str): Name of the GitHub organization.
        Returns:
            dict: Dictionary containing the organization's details from the
            GitHub API.
        """
        result = self.call_get(f"/orgs/{org}")

        if result.status_code == 200:
            org_info = result.json()

            return org_info

        elif result.status_code == 404:
            logger.info(
                f"The organization {org} was not found or there"
                " is a permission issue!"
            )

    def validate_sso(self, org: str, repository: str):
        """Query a repository in the organization to determine if SSO has been
        enabled for this PAT.

        If the query returns a 403 and an error message of "Resource protected
        by organization SAML enforcement. You must grant your Personal Access
        token access to this organization." then the PAT does not have
        permissions to this organization.

        Args:
            repository (str): Repository name in org/Repo format.
        Returns:
            bool: True if the organization is accessible either because SSO is
            not enabled, or if the PAT has been validated with SSO to that
            organization.
        """
        org_repos = self.call_get(f"/orgs/{org}/repos")

        if org_repos.status_code != 200:
            logger.warning(
                "SSO does not seem to be enabled for this PAT!"
                " Error message:"
                f" {org_repos.json()['message']}"
            )
            return False

        result = self.call_get(f"/repos/{repository}")
        if result.status_code == 403:
            logger.warning(
                "SSO does not seem to be enabled for this PAT! However,"
                "this PAT does have some access to the GitHub Enterprise. "
                f"Error message: {result.json()['message']}"
            )
            return False
        else:
            return True

    def check_org_runners(self, org: str):
        """Checks runners associated with an organization.

        This requires a token with the `admin:org` scope.

        Args:
            org (str): Name of the organization

        Returns:
            dict: Dictionary containing information about the runners.
        """
        result = self.call_get(f"/orgs/{org}/actions/runners")

        if result.status_code == 200:

            runner_info = result.json()
            if runner_info["total_count"] > 0:
                return runner_info
        else:
            logger.warning(
                f"Unable to query runners for {org}! This is likely due to the"
                " PAT permission level!"
            )

    def get_org_repo_names_graphql(self, org: str, type: str):
        """Retrieve repositories within an organization using GraphQL."""
        repo_names = []
        if type not in ["PUBLIC", "PRIVATE"]:
            raise ValueError("Unsupported type!")

        cursor = None
        while True:

            query = {
                "query": GqlQueries.GET_ORG_REPOS,
                "variables": {"orgName": org, "repoTypes": type, "cursor": cursor},
            }

            response = self.call_post("/graphql", query)
            if response.status_code == 200:
                response = response.json()
                repos = [
                    edge["node"]["name"]
                    for edge in response["data"]["organization"]["repositories"][
                        "edges"
                    ]
                ]
                repo_names.extend(repos)

                pageInfo = response["data"]["organization"]["repositories"]["pageInfo"]
                cursor = pageInfo["endCursor"] if pageInfo["hasNextPage"] else None

                if not pageInfo["hasNextPage"]:
                    break
            else:
                break

        return repo_names

    def check_org_repos(self, org: str, repo_type: str):
        """Check repositories present within an organization.

        Args:
            org (str): Organization to check repositories for.
            private (bool, optional): Whether to only check private
            repositories. Defaults to True.

        Returns:
            list: List of dictionaries representing repositories within an
            organization.
        """
        if repo_type not in [
            "all",
            "public",
            "private",
            "forks",
            "sources",
            "member",
            "internal",
        ]:
            raise ValueError("Unsupported type!")
        repos = []

        org_details = self.call_get(f"/orgs/{org}")
        # For public repos, Gato-X uses a fast GraphQL approach.
        if org_details.status_code == 200 and repo_type == "public":
            repo_count = org_details.json()["public_repos"]
            pub_repos = DataIngestor.perform_parallel_repo_ingest(self, org, repo_count)
            repos.extend([repo for repo in pub_repos if not repo["archived"]])
            return repos

        get_params = {"type": repo_type, "per_page": 100, "page": 1}

        org_repos = self.call_get(f"/orgs/{org}/repos", params=get_params)

        if org_repos.status_code == 200:
            listing = org_repos.json()

            repos.extend([repo for repo in listing if not repo["archived"]])
            # Check if there are more pages
            while len(listing) == 100:
                get_params["page"] += 1
                org_repos = self.call_get(f"/orgs/{org}/repos", params=get_params)
                if org_repos.status_code == 200:
                    listing = org_repos.json()
                    repos.extend([repo for repo in listing if not repo["archived"]])
        else:
            logger.info(f"[-] {org} requires SSO!")
            return None

        return repos

    def check_user(self):
        """Gets the authenticated user associated with a GitHub PAT and returns
        the username and available scopes.

        Format:

        {
            'user': username,
            'scopes': [ scope0, scope1, ...]
        }

        Returns:
            dict: User associated with the PAT, None otherwise.
        """
        result = self.call_get("/user")

        if result.status_code == 200:
            resp_headers = result.headers.get("x-oauth-scopes")
            if resp_headers:
                scopes = [scope.strip() for scope in resp_headers.split(",")]
            else:
                scopes = []

            user_scopes = {
                "user": result.json()["login"],
                "scopes": scopes,
                "name": result.json()["name"],
            }

            return user_scopes
        else:
            logger.warning("Provided token was not valid or has expired!")

        return None

    def get_repo_branch(self, repo: str, branch: str):
        """Check whether a specific branch exists on a remote.

        Args:
            repo (str): Name of the repository to check.
            branch (str): Name of the branch to check.

        Returns:
            int: Returns 1 upon success, 0 if the branch was not found, and -1
            if there was a failure retrieving the branch.
        """
        res = self.call_get(f"/repos/{repo}/branches/{branch}")
        if res.status_code == 200:
            return 1
        elif res.status_code == 404:
            return 0
        else:
            logger.warning("Failed to check repo for branch! " f"({res.status_code}")
            return -1

    def get_repo_runners(self, full_name: str):
        """Get self-hosted runners associated with the repository.

        Args:
            full_name (str): Name of the repository in Org/Repo format.

        Returns:
            list: List of self hosted runners from the repository.
        """
        runners = self.call_get(f"/repos/{full_name}/actions/runners")

        if runners.status_code == 200:
            runner_list = runners.json()["runners"]
            return runner_list

        return []

    def retrieve_run_logs(self, repo_name: str, workflows: list = []):
        """Retrieve the most recent run log associated with a repository.

        Args:
            repo_name (str): Full name of the repository.
            first instance of a non-ephemeral self-hosted runner is detected.
            Defaults to True.
            workflows (list, optional): List of workflows to check for. Defaults
            to empty list.
        Returns:
            list: List of run logs for runs that ran on self-hosted runners.
        """
        start_date = datetime.now() - timedelta(days=60)
        runs = []

        for workflow in workflows:
            # Get workflow runs for workflows we think have a sh runner.
            run_result = self.call_get(
                f"/repos/{repo_name}/actions/workflows/{workflow}/runs",
                params={
                    "per_page": "3",
                    "status": "completed",
                    "exclude_pull_requests": "true",
                    "created": f">{start_date.isoformat()}",
                },
            )

            if run_result.status_code == 200:
                runs.extend(run_result.json()["workflow_runs"])

        # This is a dictionary so we can de-duplicate runner IDs based on
        # the machine_name:runner_name.
        run_logs = {}
        names = set()
        total_attempts = 0

        if runs:
            logger.debug(f"Enumerating runs within {repo_name}")
        for run in runs:
            # We only look at 10 workflow logs.
            # If we haven't found a non-ephemeral runner it is unlikely we will.
            # Larger repos with complex matrix builds and reusable workflows
            # can have massive log sizes and we end up wasting a lot of time.
            if total_attempts > 10:

                break

            # We are only interested in runs that actually executed.
            if run["conclusion"] != "success" and run["conclusion"] != "failure":
                continue

            # We only look at one workflow run (for yaml) per branch
            workflow_key = f"{run['head_branch']}:{run['path']}"
            if workflow_key in names:
                continue
            names.add(workflow_key)
            run_log = self.call_get(
                f'/repos/{repo_name}/actions/runs/{run["id"]}/'
                f'attempts/{run["run_attempt"]}/logs'
            )

            if run_log.status_code == 200:
                try:
                    run_log = self.__process_run_log(run_log.content, run)
                    if run_log:
                        key = f"{run_log['machine_name']}:{run_log['runner_name']}"
                        run_logs[key] = run_log

                        if run_log["non_ephemeral"]:
                            return run_logs.values()
                except Exception as e:
                    logger.warning(
                        f"Failed to process run log for {repo_name} run "
                        f"{run['id']} attempt {run['run_attempt']}!"
                    )
            elif run_log.status_code == 410:
                break
            else:
                logger.debug(
                    f"Call to retrieve run logs from {repo_name} run "
                    f"{run['id']} attempt {run['run_attempt']} returned "
                    f"{run_log.status_code}!"
                )

            total_attempts += 1

        return run_logs.values()

    def parse_workflow_runs(self, repo_name: str):
        """Returns the number of workflow runs associated with the repository.

        Args:
            repo_name (str): Name of the repository in Org/Repo format to parse
            workflow runs for.

        Returns:
            int: Number of workflow runs associated with the repository, None
            if there was a failure.
        """
        runs = self.call_get(f"/repos/{repo_name}/actions/runs")

        if runs.status_code == 200:

            return runs.json()["total_count"]
        else:
            logger.warning("Unable to query workflow runs.")

        return None

    def get_recent_workflow(
        self, repo_name: str, sha: str, file_name: str, time_after=None
    ) -> int:
        """
        This function is used to get the most recent workflow from a GitHub repository.

        Parameters:
            repo_name (str): The name of the repository. It should be in the format 'owner/repo'.
            sha (str): The SHA of the commit for which to get the workflow.
            file_name (str): The name of the workflow file (without the .yml extension).
            time_after (str, optional): A timestamp in ISO 8601 format:
            YYYY-MM-DDTHH:MM:SSZ. Only show workflows updated after this time.

        Returns:
            int: The ID of the workflow if found, 0 if no workflows are found, or -1
            if there was an error querying the workflows.

        Raises:
            None

        Example:
            get_recent_workflow(
            'octocat/Hello-World',
            '7fd1a60b01f91b314f59955a4e4d4e80d8edf11d', 'test_workflow'
        )
        """
        params = {"head_sha": sha}

        if time_after:
            params["created"] = time_after

        req = self.call_get(f"/repos/{repo_name}/actions/runs", params=params)

        if req.status_code != 200:
            logger.warning("Unable to query workflow runs.")
            return -1

        data = req.json()

        if data["total_count"] == 0:
            return 0

        # Find the id of the workflow
        for workflow in data["workflow_runs"]:
            if f".github/workflows/{file_name}.yml" in workflow["path"]:
                return workflow["id"]

        return 0

    def get_workflow_status(self, repo_name: str, workflow_id: int):
        """Returns the status if the workflow by id.

        Args:
            repo_name (str): Name of the repository that has the workflow.
            workflow_id (int): ID of the workflow.

        Returns:
            int: 1 if the workflow has completed, 0 if it is pending, and -1 if
            there was a failure.
        """
        req = self.call_get(f"/repos/{repo_name}/actions/runs/{workflow_id}")

        if req.status_code != 200:
            logger.warning("Unable to query the workflow.")
            return -1

        data = req.json()

        if data.get("status", "queued") in ["queued", "in_progress"]:
            return 0
        return 1 if data.get("conclusion", "failure") == "success" else -1

    def delete_workflow_run(self, repo_name: str, workflow_id: int):
        """Deletes a previous workflow run.

        Args:
            repo_name (str): Name of the repository that has the workflow.
            workflow_id (int): ID of the workflow.

        Returns:
            bool: True if the workflow was deleted, false otherwise.
        """
        req = self.call_delete(f"/repos/{repo_name}/actions/runs/" f"{workflow_id}")

        return req.status_code == 204

    def download_workflow_logs(self, repo_name: str, workflow_id: int):
        """Download worfklow run logs and saves them to a zip file under the
        workflow ID.

        Args:
            repo_name (str): Name of the repository that has the workflow.
            workflow_id (int): ID of the workflow.

        Returns:
            bool: True of the workflow log was downloaded, false otherwise.
        """
        req = self.call_get(f"/repos/{repo_name}/actions/runs/" f"{workflow_id}/logs")

        if req.status_code != 200:
            return False

        with open(f"{workflow_id}.zip", "wb+") as f:
            f.write(req.content)
        return True

    def retrieve_workflow_log(self, repo_name: str, workflow_id: int, job_name: str):
        """Download single run log and returns the text output from the zip.

        Args:
            repo_name (str): Name of the repository that has the workflow.
            workflow_id (int): ID of the workflow.
            job_name (str): Name of job to get output from.
        Returns:
            str: String content of the run log matching the job name, if found.
        """
        req = self.call_get(f"/repos/{repo_name}/actions/runs/" f"{workflow_id}/logs")

        if req.status_code != 200:
            return False

        return self.__get_full_runlog(req.content, job_name)

    def retrieve_workflow_artifact(self, repo_name: str, workflow_id: int):
        """Download workflow artifacts and return the files. Only use this for
        small artifacts, as this extracts the zip in memory.

        Args:
            repo_name (str): Name of the repository that has the workflow.
            workflow_id (int): ID of the workflow.
        Returns:

        """
        files = {}

        req = self.call_get(
            f"/repos/{repo_name}/actions/runs/" f"{workflow_id}/artifacts"
        )
        if req.status_code != 200:
            return False

        artifacts = req.json().get("artifacts", [])

        if artifacts:
            download_url = artifacts[0]["archive_download_url"]

            archive = self.call_get(download_url.replace("https://api.github.com", ""))

            with zipfile.ZipFile(io.BytesIO(archive.content)) as artifact:
                for zipinfo in artifact.infolist():
                    with artifact.open(zipinfo) as run_log:
                        content = run_log.read()
                        files[zipinfo.filename] = content

        return files

    def download_workflow_artifact(
        self, repo_name: str, workflow_id: int, destination: str
    ):
        """Download a workflow artifact and save it to the destination."""

        req = self.call_get(
            f"/repos/{repo_name}/actions/runs/" f"{workflow_id}/artifacts"
        )
        if req.status_code != 200:
            return False

        artifacts = req.json().get("artifacts", [])
        download_url = artifacts[0]["archive_download_url"]

        archive = self.call_get(download_url.replace("https://api.github.com", ""))

        with open(destination, "wb") as f:
            f.write(archive.content)

            return destination

        return False

    def create_branch(self, repo_name: str, branch_name: str):
        """Create a branch with the provided name.

        Args:
            repo_name (str): Name of repository in Org/Repo format.
            branch_name (str): Name of branch to create.
        """
        resp = self.call_get(f"/repos/{repo_name}")
        default_branch = resp.json()["default_branch"]
        resp = self.call_get(f"/repos/{repo_name}/git/ref/heads/{default_branch}")

        json_resp = resp.json()
        sha = json_resp["object"]["sha"]

        branch_data = {"ref": f"refs/heads/{branch_name}", "sha": sha}

        resp = self.call_post(f"/repos/{repo_name}/git/refs", params=branch_data)

        if resp.status_code == 201:
            return True
        else:
            return False

    def delete_branch(self, repo_name: str, branch_name: str):
        """Deletes the specified branch within the repository.

        Args:
            repo_name (str): Name of the repository in Owner/Repo format.
            branch_name (str): Name of the branch to delete.
        """
        resp = self.call_delete(f"/repos/{repo_name}/git/refs/heads/{branch_name}")

        if resp.status_code == 204:
            return True

    def commit_file(
        self,
        repo_name: str,
        branch_name: str,
        file_path: str,
        file_content: bytes,
        commit_author: str = "Gato-X",
        commit_email: str = "gato-x@pwn.com",
        message="Testing",
    ):
        """Commits a file to the specified branch on a repository.

        Args:
            repo_name (str): Name of repository to target.
            branch_name (str): Branch name to commit to. Must exist, otherwise
            the operation will fail.
            file_path (str): Path within to repository to commit file to.
            file_content (bytes): Content of the file to commit in bytes.
            commit_author (str): Author of the commit.
            commit_email (str): Email for commit.
            message (str): Commit message for testing.
        """
        b64_contents = base64.b64encode(file_content)
        commit_data = {
            "message": message,
            "content": b64_contents.decode("utf-8"),
            "branch": branch_name,
            "committer": {"name": commit_author, "email": commit_email},
        }

        resp = self.call_put(
            f"/repos/{repo_name}/contents/{file_path}", params=commit_data
        )

        if resp.status_code == 201:
            resp_json = resp.json()
            return resp_json["commit"]["sha"]
        else:
            print(resp.status_code)
            print(resp.text)

    def retrieve_workflow_ymls(self, repo_name: str):
        """Retrieve all .yml or .yaml files within the workflows directory.
        Utilizes the GitHub Repository contents API.

        Args:
            repo_name (str): Name of the repository in Org/Repo format.

        Returns:
            (list): List of yml files in text format.
        """
        ymls = []

        resp = self.call_get(f"/repos/{repo_name}/contents/.github/workflows/")

        if resp.status_code == 200:
            objects = resp.json()

            for file in objects:
                if file["type"] == "file" and (
                    file["name"].endswith(".yml") or file["name"].endswith(".yaml")
                ):

                    resp = self.call_get(f'/repos/{repo_name}/contents/{file["path"]}')
                    if resp.status_code == 200:
                        resp_data = resp.json()
                        if "content" in resp_data:
                            file_data = base64.b64decode(resp_data["content"])
                            ymls.append(Workflow(repo_name, file_data, file["name"]))

        return ymls

    def retrieve_repo_file(
        self, repo_name: str, file_path: str, ref: str, public=False
    ):
        """Retrieves a single file from a GitHub repository.

        If the repository is public, it instead uses the raw.githubusercontent.com
        endpoint to save on the API rate limit.
        """
        file_data = None

        if public:
            file_data = self.__get_raw_file(repo_name, file_path, ref)
        else:
            resp = self.call_get(
                f"/repos/{repo_name}/contents/{file_path}", params={"ref": ref}
            )
            if resp.status_code == 200:
                resp_data = resp.json()
                if "content" in resp_data:
                    file_data = base64.b64decode(resp_data["content"])

        if file_data:
            return Workflow(
                repo_name,
                file_data,
                file_path.rsplit("/", 1)[-1],
                non_default=ref,
                special_path=file_path,
            )

    def retrieve_workflow_yml(self, repo_name: str, workflow_name: str):
        """Retrieve all .yml or .yaml files within the workflows directory.
        Utilizes the GitHub Repository contents API.

        Args:
            repo_name (str): Name of the repository in Org/Repo format.
            workflow_name (str): Name of the workflow

        Returns:
            (list): List of yml files in text format.
        """
        resp = self.call_get(
            f"/repos/{repo_name}/contents/.github/workflows/{workflow_name}"
        )

        if resp.status_code == 200:

            resp_data = resp.json()
            if "content" in resp_data:
                file_data = base64.b64decode(resp_data["content"])
                return Workflow(repo_name, file_data, workflow_name)
        else:
            raise ValueError(
                f"Failed to retrieve workflow {workflow_name} from {repo_name}!"
            )

    def get_secrets(self, repo_name: str):
        """Issues an API call to the GitHub API to list secrets for a
        repository. This will succeed as long as the token has the repo scope
        and the user has write access to the repository.

        Args:
            repo_name (str): Name of repository to list secrets for.
        Returns:
            (list): List of secrets at the repo level, empty list if none.
        """
        secrets = []

        resp = self.call_get(f"/repos/{repo_name}/actions/secrets")
        if resp.status_code == 200:
            secrets_response = resp.json()

            if secrets_response["total_count"] > 0:
                secrets = secrets_response["secrets"]

        return secrets

    def get_environment_secrets(self, repo_name: str, environment_name: str):
        """Issues an API call to the GitHub API to list secrets for a specific
        environment within a repository. This requires the token to have the repo
        scope and the user to have write access to the repository and environment.

        Args:
            repo_name (str): Name of the repository.
            environment_name (str): Name of the environment to list secrets for.

        Returns:
            list: List of secrets at the environment level, empty list if none.
        """
        secrets = []

        environment_name = environment_name.replace("/", "%2F")
        resp = self.call_get(
            f"/repos/{repo_name}/environments/{environment_name}/secrets"
        )
        if resp.status_code == 200:
            secrets_response = resp.json()

            if secrets_response["total_count"] > 0:
                secrets = secrets_response["secrets"]

        return secrets

    def get_org_secrets(self, org_name: str):
        secrets = []

        resp = self.call_get(f"/orgs/{org_name}/actions/secrets")
        if resp.status_code == 200:
            secrets_response = resp.json()

            if secrets_response["total_count"] > 0:
                for secret in secrets_response["secrets"]:

                    if secret["visibility"] == "selected":

                        repos_resp = self.call_get(
                            f"/orgs/{org_name}/actions/secrets/"
                            f'{secret["name"]}/repositories'
                        )

                        if repos_resp.status_code == 200:
                            repos_json = repos_resp.json()
                            repo_names = [
                                repo["full_name"] for repo in repos_json["repositories"]
                            ]

                        secret["repos"] = repo_names

                    secrets.append(secret)

        return secrets

    def get_repo_org_secrets(self, repo_name: str):
        """Issues an API call to the GitHub API to list org secrets for a
        repository. This will succeed as long as the token has the repo scope
        and the user has write access to the repository.

        Args:
            repo_name (str): Name of repository to list secrets for.

        Returns:
            (list): List of org secrets that can be read via a workflow in this
            repository.
        """
        resp = self.call_get(f"/repos/{repo_name}/actions/organization-secrets")
        secrets = []
        if resp.status_code == 200:
            secrets_response = resp.json()

            if secrets_response["total_count"] > 0:
                secrets = secrets_response["secrets"]

        return secrets

    def get_file_last_updated(self, repo_name: str, file_path: str):
        resp = self.call_get(
            f"/repos/{repo_name}/commits", params={"path": file_path, "per_page": 1}
        )

        commit_date = resp.json()[0]["commit"]["author"]["date"]
        commit_author = resp.json()[0]["commit"]["author"]["name"]
        commit_sha = resp.json()[0]["sha"]

        return commit_date, commit_author, commit_sha

    def get_all_environment_protection_rules(self, repo_name: str):
        """
        Query all environments for a GitHub repository and return the combined protection rules array.

        Args:
            repo_name (str): The name of the repository.

        Returns:
            list: The combined protection rules array from all environments.
        """
        response = self.call_get(f"/repos/{repo_name}/environments")

        all_protection_rules = []

        if response.status_code == 200:
            all_environments = response.json()

            for environment in all_environments["environments"]:
                protection_rules = environment.get("protection_rules", [])
                all_protection_rules.extend(
                    [
                        environment["name"]
                        for rule in protection_rules
                        if rule["type"] == "required_reviewers"
                    ]
                )

        return all_protection_rules

    def commit_workflow(
        self,
        repo_name: str,
        target_branch: str,
        workflow_contents: bytes,
        file_name: str,
        commit_author: str = "Gato-X",
        commit_email: str = "Gato-X@pwn.com",
        message="Testing",
    ):
        """
        Commits a new workflow file to a specified repository.

        This function performs the following steps:
        1. Gets the latest commit SHA of the target branch.
        2. Gets the tree SHA of the latest commit of the new branch.
        3. Gets the tree of the .github/workflows directory.
        4. If the workflows tree exists, it gets the SHA of the workflows tree.
        5. Creates a new tree where all blobs in the .github/workflows tree are removed.
        6. Creates a new commit on the new branch with the new tree.
        7. Updates the new branch to point to the new commit.

        Args:
            repo_name (str): The name of the repository.
            target_branch (str): The name of the target branch.
            workflow_contents (bytes): The content of the new workflow file.
            file_name (str): The name of the new workflow file.
            commit_author (str, optional): The author of the commit. Defaults to "Gato".
            commit_email (str, optional): The email of the commit author. Defaults to "gato@gato.infosec".
            message (str, optional): The commit message. Defaults to "Testing".

        Returns:
            str: The SHA of the new commit if the commit was successful, None otherwise.
        """
        # Step 1: Get latest commit SHA of target branch
        r = self.call_get(f"/repos/{repo_name}")
        if self.__verify_result(r, 200) is False:
            return None
        default_branch = r.json()["default_branch"]

        r = self.call_get(f"/repos/{repo_name}/commits/{default_branch}")
        if self.__verify_result(r, 200) is False:
            return None
        latest_commit_sha = r.json()["sha"]

        # Step 2: Get tree SHA of latest commit of default
        r = self.call_get(f"/repos/{repo_name}/git/commits/{latest_commit_sha}")
        if self.__verify_result(r, 200) is False:
            return None
        tree_sha = r.json()["tree"]["sha"]

        # Step 3: Get the tree of the .github/workflows directory
        r = self.call_get(
            f"/repos/{repo_name}/git/trees/{tree_sha}", params={"recursive": "1"}
        )
        if self.__verify_result(r, 200) is False:
            return None

        base_sha = r.json()["sha"]
        tree = r.json()["tree"]

        existing_files = (
            item
            for item in tree
            if ".github/workflows" in item["path"] and item["type"] == "blob"
        )

        # Step 4: Create a new tree where all blobs in the .github/workflows
        # tree are removed
        new_workflow_file_content = base64.b64encode(workflow_contents).decode()

        r = self.call_post(
            f"/repos/{repo_name}/git/blobs",
            params={"content": new_workflow_file_content, "encoding": "base64"},
        )
        if self.__verify_result(r, 201) is False:
            return None

        new_tree = [
            {
                "path": f".github/workflows/{file_name}",
                "mode": "100644",
                "type": "blob",
                "sha": r.json()["sha"],
            }
        ]

        # Delete everything else
        for existing in existing_files:
            # Don't delete the same file - this will happen if the workflow
            # already exists (such as a test.yml file)
            if existing["path"] == f".github/workflows/{file_name}":
                continue

            new_tree.append(
                {
                    "path": existing["path"],
                    "mode": existing["mode"],
                    "type": existing["type"],
                    "sha": None,
                }
            )

        r = self.call_post(
            f"/repos/{repo_name}/git/trees",
            params={"base_tree": base_sha, "tree": new_tree},
        )
        if self.__verify_result(r, 201) is False:
            return None
        new_tree_sha = r.json()["sha"]

        # Step 5: Create new commit on new branch
        r = self.call_post(
            f"/repos/{repo_name}/git/commits",
            params={
                "message": message,
                "tree": new_tree_sha,
                "parents": [latest_commit_sha],
                "author": {"name": commit_author, "email": commit_email},
            },
        )
        new_commit_sha = r.json()["sha"]

        # Step 6: Update the new branch to point to the new commit
        r = self.call_post(
            f"/repos/{repo_name}/git/refs",
            params={"sha": new_commit_sha, "ref": f"refs/heads/{target_branch}"},
        )
        if self.__verify_result(r, 201) is False:
            return None

        return new_commit_sha

    def backtrack_head(self, repo_name, ref_name, commit_depth):
        """Uses the Git database API to revert a number of commits back from head.

        This essentially does:

            git reset --hard HEAD~<COMMIT_DEPTH>
            git push --force

        This is used for force pushing off payloads when conducting attacks
        in order to close the pull request.
        """

        params = {"sha": ref_name, "per_page": commit_depth + 1}

        resp = self.call_get(f"/repos/{repo_name}/commits", params=params)

        if resp.status_code == 200:
            commits = resp.json()
            target = commits[commit_depth]["sha"]
        else:
            return False

        resp = self.call_patch(
            f"/repos/{repo_name}/git/refs/heads/{ref_name}",
            params={"sha": target, "force": True},
        )

        if resp.status_code == 200:
            return True
        else:
            return False

    def issue_dispatch(
        self, repo_name, target_workflow, target_branch, dispatch_inputs
    ):
        """Issues a workflow dispatch event to trigger a workflow.

        Args:
            repo_name (str): Name of the repository in Org/Repo format.
            target_workflow (str): Name of the workflow to trigger.
        """
        r = self.call_post(
            f"/repos/{repo_name}/actions/workflows/{target_workflow}/dispatches",
            params={"ref": target_branch, "inputs": dispatch_inputs},
        )

        return r.status_code == 204

    def get_issue_comments(self, repo_name, target_pr):
        """Receives the last 5 comments on the PR within the last minute.

        Args:
            repo_name (str): Name of the repository in Org/Repo format.
            target_pr (int): PR number to get comments for.
        """

        since = (
            (datetime.now(datetime.UTC) - timedelta(minutes=1))
            .replace(microsecond=0)
            .isoformat()
        )
        params = {"per_page": 5, "since": since + "Z"}

        r = self.call_get(
            f"/repos/{repo_name}/issues/{target_pr}/comments", params=params
        )

        return r.json()

    def create_repository(self, repository_name: str):
        """Creates a private repository for the authenticated user."""

        params = {"private": True, "name": repository_name}

        response = self.call_post(f"/user/repos", params=params)

        if response.status_code == 201:
            return response.json()["full_name"]
        else:
            return False

    def create_pull_request(
        self,
        source_repo: str,
        source_branch: str,
        target_repo: str,
        target_banch: str,
        pr_body="",
        pr_title="CI Test",
        draft=True,
    ):
        """
        This function is used to create a pull request on GitHub.

        Parameters:
            source_repo (str): The name of the source repository. It should be in the format 'owner/repo'.
            source_branch (str): The name of the source branch.
            target_repo (str): The name of the target repository. It should be in the format 'owner/repo'.
            target_banch (str): The name of the target branch.
            pr_body (str, optional): The body of the pull request. Default is an empty string.
            pr_title (str, optional): The title of the pull request. Default is 'CI Test'.
            draft (bool, optional): Whether the pull request should be created as a draft. Default is True.

        Returns:
            str: The URL of the created pull request if successful, False otherwise.

        Raises:
            None

        Example:
            create_pull_request('octocat/Hello-World', 'feature-branch', 'octocat/Hello-World', 'main', pr_body='This is a test PR', pr_title='Test PR', draft=False)
        """

        params = {
            "title": pr_title,
            "body": pr_body,
            "head": source_branch,
            "base": target_banch,
            "head_repo": source_repo,
            "draft": draft,
        }

        response = self.call_post(f"/repos/{target_repo}/pulls", params=params)

        if response.status_code == 201:
            return response.json()["html_url"]
        else:
            return False

    def retrieve_raw_action(self, repo: str, file_path: str, ref: str):
        """Retrieves a GitHub action yaml file from a public repository."""
        if file_path.endswith(".yml") or file_path.endswith(".yaml"):
            file_path = file_path.replace("//", "/")
            paths = [file_path]
        else:
            if not file_path.endswith("/"):
                file_path += "/"
            elif file_path.endswith("//"):
                file_path = file_path.replace("//", "/")
            paths = [f"{file_path}action.yml", f"{file_path}action.yaml"]

        for path in paths:
            res = self.__get_raw_file(repo, path, ref)

            if res:
                return res

        return None

    def get_installation_repos(self):
        """ """
        response = self.call_get("/installation/repositories")
        if response.status_code == 200:
            return response.json()

    def get_commit_merge_date(self, repo: str, sha: str):
        """Gets the date of the merge commit."""

        query = {
            "query": GqlQueries.GET_PR_MERGED,
            "variables": {
                "sha": sha,
                "repo": repo.split("/")[1],
                "owner": repo.split("/")[0],
            },
        }

        r = self.call_post("/graphql", params=query)
        if r.status_code == 200:
            response = r.json()

            if not response["data"]["repository"]:
                return None

            if not response["data"]["repository"]["commit"]["associatedPullRequests"][
                "edges"
            ]:
                return None

            pr_info = response["data"]["repository"]["commit"][
                "associatedPullRequests"
            ]["edges"][0]["node"]

            if pr_info["merged"]:
                return pr_info["mergedAt"]



"""
Copyright 2025, Adnan Khan

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from gatox.workflow_graph.nodes.step import StepNode
from gatox.workflow_graph.nodes.workflow import WorkflowNode
from gatox.workflow_graph.nodes.job import JobNode
from gatox.workflow_graph.nodes.repo import RepoNode
from gatox.workflow_graph.nodes.action import ActionNode
from gatox.models.workflow import Workflow
from gatox.models.repository import Repository
from gatox.workflow_parser.utility import parse_github_path


class NodeFactory:
    """
    A factory class for creating various types of nodes and caching them.

    Attributes:
        NODE_CACHE (dict): A cache to store created nodes by their names.
    """

    NODE_CACHE = {}

    @staticmethod
    def create_repo_node(repo_wrapper: Repository):
        """
        Create a RepoNode for the given repository and cache it.

        This method checks if a RepoNode for the specified repository already exists
        in the cache. If it does, it returns the cached node. Otherwise, it creates
        a new RepoNode, caches it, and then returns it.

        Args:
            repo_wrapper (Repository): The repository wrapper containing repository details.

        Returns:
            tuple: A tuple containing the RepoNode instance and a boolean indicating
                   whether the node was newly created (True) or retrieved from the cache (False).
        """
        if repo_wrapper.name in NodeFactory.NODE_CACHE:
            return NodeFactory.NODE_CACHE[repo_wrapper.name], False
        else:
            repo_node = RepoNode(repo_wrapper)
            NodeFactory.NODE_CACHE[repo_node.name] = repo_node
            return repo_node, True

    @staticmethod
    def create_job_node(job_name, ref, repo_name, workflow_path, needs: list = []):
        """
        Create a JobNode for the specified job and cache it.

        This method checks if a JobNode with the given name already exists in the cache.
        If it does, it returns the cached node. Otherwise, it creates a new JobNode,
        caches it, and then returns it.

        Args:
            job_name (str): The name of the job.
            ref (str): The reference (e.g., branch or tag) associated with the job.
            repo_name (str): The name of the repository where the job resides.
            workflow_path (str): The path to the workflow file containing the job.
            needs (list, optional): A list of dependencies required by the job.

        Returns:
            JobNode: The created or cached JobNode instance.
        """
        job_node = JobNode(job_name, ref, repo_name, workflow_path)
        if job_node.name in NodeFactory.NODE_CACHE:
            return NodeFactory.NODE_CACHE[job_node.name]
        else:
            NodeFactory.NODE_CACHE[job_node.name] = job_node
            return job_node

    @staticmethod
    def create_workflow_node(workflow_data: Workflow, ref, repo_name, workflow_path):
        """
        Create a WorkflowNode for the specified workflow and cache it.

        This method checks if a WorkflowNode with the given name already exists in the cache.
        If it does, it initializes the existing node with the provided workflow data.
        Otherwise, it creates a new WorkflowNode, initializes it with the workflow data,
        caches it, and then returns it.

        Args:
            workflow_data (Workflow): The workflow wrapper containing workflow details.
            ref (str): The reference (e.g., branch or tag) associated with the workflow.
            repo_name (str): The name of the repository where the workflow resides.
            workflow_path (str): The path to the workflow file.

        Returns:
            WorkflowNode: The created or initialized WorkflowNode instance.
        """
        workflow_node = WorkflowNode(ref, repo_name, workflow_path)
        if workflow_node.name in NodeFactory.NODE_CACHE:
            NodeFactory.NODE_CACHE[workflow_node.name].initialize(workflow_data)
            return NodeFactory.NODE_CACHE[workflow_node.name]
        else:
            NodeFactory.NODE_CACHE[workflow_node.name] = workflow_node
            workflow_node.initialize(workflow_data)
            return workflow_node

    @staticmethod
    def create_called_workflow_node(callee: str, caller_ref, caller_repo):
        """
        Create a WorkflowNode for a called workflow and cache it.

        This method parses the callee string to determine the workflow's repository,
        path, and reference. It then checks if a WorkflowNode with the constructed name
        exists in the cache. If it does, it returns the cached node. Otherwise, it creates
        a new WorkflowNode, caches it, and then returns it.

        Args:
            callee (str): The reference to the called workflow, which could be a relative path
                          (starting with "./") or a full GitHub path with a specific ref.
            caller_ref (str): The reference (e.g., branch or tag) of the calling workflow.
            caller_repo (str): The name of the repository where the calling workflow resides.

        Returns:
            WorkflowNode: The created or cached WorkflowNode instance.
        """
        if callee.startswith("./"):
            workflow_name = callee.split("/")[-1]
            workflow_path = f".github/workflows/{workflow_name}"
            repo_name = caller_repo
            ref = caller_ref
        elif "@" in callee:
            repo_name, path, ref = parse_github_path(callee)
            workflow_name = path.split("/")[-1]
            workflow_path = f".github/workflows/{workflow_name}"

        name = f"{repo_name}:{ref}:{workflow_path}"
        if name in NodeFactory.NODE_CACHE:
            return NodeFactory.NODE_CACHE[name]
        else:
            workflow_node = WorkflowNode(ref, repo_name, workflow_path)
            NodeFactory.NODE_CACHE[workflow_node.name] = workflow_node
            return workflow_node

    @staticmethod
    def create_step_node(
        step_data, ref, repo_name, workflow_path, job_name, step_number
    ):
        """
        Create a StepNode for the specified step and cache it.

        This method creates a new StepNode with the provided data and caches it using its
        name as the key. It then returns the created StepNode instance.

        Args:
            step_data (dict): The data dictionary containing step details.
            ref (str): The reference (e.g., branch or tag) associated with the step.
            repo_name (str): The name of the repository where the step resides.
            workflow_path (str): The path to the workflow file containing the step.
            job_name (str): The name of the job that includes this step.
            step_number (int): The sequential number of the step within the job.

        Returns:
            StepNode: The created StepNode instance.
        """
        step_node = StepNode(
            step_data, ref, repo_name, workflow_path, job_name, step_number
        )
        NodeFactory.NODE_CACHE[step_node.name] = step_node
        return step_node

    @staticmethod
    def create_action_node(action_name, ref, action_path, repo_name, params={}):
        """
        Create an ActionNode for the specified action and cache it.

        This method constructs the name of the ActionNode using the repository name,
        reference, action path, and action name. It then checks if an ActionNode with
        this name exists in the cache. If it does, it returns the cached node.
        Otherwise, it creates a new ActionNode, caches it, and then returns it.

        Args:
            action_name (str): The name of the action.
            ref (str): The reference (e.g., branch or tag) associated with the action.
            action_path (str): The file system path to the action within the repository.
            repo_name (str): The name of the repository where the action resides.
            params (dict, optional): Additional parameters for the action.

        Returns:
            ActionNode: The created or cached ActionNode instance.
        """
        name = f"{repo_name}:{ref}:{action_path}:{action_name}"
        if name in NodeFactory.NODE_CACHE:
            return NodeFactory.NODE_CACHE[name]
        else:
            action_node = ActionNode(action_name, ref, action_path, repo_name, params)
            NodeFactory.NODE_CACHE[action_node.name] = action_node
            return action_node



"""
Copyright 2025, Adnan Khan

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from abc import abstractmethod
from gatox.enumerate.results.confidence import Confidence
from gatox.enumerate.results.complexity import Complexity


class AnalysisResult:
    """
    Holds the results from graph traversals defined in the workflow_graph/visitors module.

    Attributes:
        repository_name (str): The name of the repository.
        issue_type (str): The type of issue identified.
        confidence_score (str): The confidence score (LOW, MEDIUM, HIGH).
        attack_complexity (str): The attack complexity score (LOW, MEDIUM, HIGH).
    """

    def __init__(
        self,
        repository_name,
        issue_type,
        confidence_score: Confidence,
        attack_complexity: Complexity,
    ):
        self.__repository_name = repository_name
        self.__confidence_score = confidence_score
        self.__issue_type = issue_type
        self.__attack_complexity = attack_complexity

    def collect_steps(self, path: list):
        for node in path:
            yield node.get_repr()

    def repo_name(self):
        """
        Gets the repository name.

        Returns:
            str: The repository name.
        """
        return self.__repository_name

    @abstractmethod
    def issue_type(self):
        """
        Gets the issue type.

        Returns:
            str: The issue type.
        """
        return self.__issue_type

    def confidence_score(self):
        """
        Gets the confidence score.

        Returns:
            str: The confidence score.
        """
        return self.__confidence_score

    def attack_complexity(self):
        """
        Gets the attack complexity score.

        Returns:
            str: The attack complexity score.
        """
        return self.__attack_complexity

    @abstractmethod
    def to_machine(self):
        pass



from unittest.mock import patch
from unittest.mock import patch, MagicMock
from gatox.workflow_parser.utility import (
    decompose_action_ref,
    parse_github_path,
    check_sinks,
    parse_script,
    starts_with_any,
    process_matrix,
    getTokens,
    check_sus,
    checkUnsafe,
    process_runner,
)


def test_get_tokens_empty():
    assert getTokens("") == []


def test_get_tokens_no_context():
    assert getTokens("plain text") == []


def test_get_tokens_single_context():
    assert getTokens("${{ github.token }}") == ["github.token"]


def test_get_tokens_multiple_contexts():
    result = getTokens("${{ github.token }} ${{ github.actor }}")
    assert "github.token" in result
    assert "github.actor" in result


def test_get_tokens_with_or():
    result = getTokens("${{ github.token || github.actor }}")
    assert "github.token" in result
    assert "github.actor" in result


def test_starts_with_any_empty():
    assert not starts_with_any("", ["test"])


def test_starts_with_any_no_match():
    assert not starts_with_any("test", ["xyz", "abc"])


def test_starts_with_any_single_match():
    assert starts_with_any("test123", ["test"])


def test_starts_with_any_multiple_matches():
    assert starts_with_any("test123", ["xyz", "test", "abc"])


@patch("gatox.workflow_parser.utility.ConfigurationManager")
def test_process_matrix_no_key(mock_config):
    job_def = {"strategy": {"matrix": {}}}
    runs_on = "${{ matrix.os }}"
    assert not process_matrix(job_def, runs_on)


@patch("gatox.workflow_parser.utility.ConfigurationManager")
def test_process_matrix_with_key(mock_config):
    mock_config().WORKFLOW_PARSING = {"GITHUB_HOSTED_LABELS": ["ubuntu-latest"]}
    job_def = {"strategy": {"matrix": {"os": ["self-hosted"]}}}
    runs_on = "${{ matrix.os }}"
    assert process_matrix(job_def, runs_on)


@patch("gatox.workflow_parser.utility.ConfigurationManager")
def test_process_matrix_includes(mock_config):
    mock_config().WORKFLOW_PARSING = {"GITHUB_HOSTED_LABELS": ["ubuntu-latest"]}
    job_def = {"strategy": {"matrix": {"include": [{"os": "self-hosted"}]}}}
    runs_on = "${{ matrix.os }}"
    assert process_matrix(job_def, runs_on)


@patch("gatox.workflow_parser.utility.ConfigurationManager")
def test_process_matrix_invalid(mock_config):
    job_def = {"strategy": "invalid"}
    runs_on = "${{ matrix.os }}"
    assert not process_matrix(job_def, runs_on)


@patch("gatox.workflow_parser.utility.ConfigurationManager")
@patch("gatox.workflow_parser.utility.pattern")
@patch("gatox.workflow_parser.utility.check_sinks")
def test_parse_script_empty(mock_check_sinks, mock_pattern, mock_config):
    mock_check_sinks.return_value = False
    result = parse_script("")
    assert result == {
        "is_checkout": False,
        "metadata": None,
        "is_sink": False,
        "hard_gate": False,
        "soft_gate": False,
    }


@patch("gatox.workflow_parser.utility.ConfigurationManager")
@patch("gatox.workflow_parser.utility.pattern")
@patch("gatox.workflow_parser.utility.check_sinks")
def test_parse_script_checkout(mock_check_sinks, mock_pattern, mock_config):
    mock_check_sinks.return_value = False
    mock_pattern.search.return_value = MagicMock(
        group=lambda x: "prBranch" if x == 2 else ""
    )
    mock_config().WORKFLOW_PARSING = {"PR_ISH_VALUES": ["pr", "pull_request"]}

    script = "git checkout prBranch"
    result = parse_script(script)
    assert result["is_checkout"] is True
    assert result["metadata"] == "prBranch"


@patch("gatox.workflow_parser.utility.ConfigurationManager")
@patch("gatox.workflow_parser.utility.pattern")
@patch("gatox.workflow_parser.utility.check_sinks")
def test_parse_script_gates_and_sinks(mock_check_sinks, mock_pattern, mock_config):
    mock_check_sinks.return_value = True
    mock_config().WORKFLOW_PARSING = {"PR_ISH_VALUES": []}
    mock_pattern.search.return_value = None

    script = (
        "some commands\n"
        "isCrossRepository is set, use GITHUB_OUTPUT\n"
        "github.rest.repos.checkCollaborator\n"
        "getCollaboratorPermissionLevel\n"
    )
    result = parse_script(script)

    assert result["hard_gate"] is True
    assert result["soft_gate"] is True
    assert result["is_sink"] is True


def test_decompose_action_ref_standard_action():
    action_path = "actions/checkout@v2"
    repo_name = "test/repo"
    result = decompose_action_ref(action_path, repo_name)
    expected = {
        "key": "actions/checkout@v2",
        "path": "",
        "ref": "v2",
        "local": False,
        "docker": False,
        "repo": "actions/checkout",
    }
    assert result == expected


def test_decompose_action_ref_local_action():
    action_path = "./local-action"
    repo_name = "test/repo"
    result = decompose_action_ref(action_path, repo_name)
    expected = {
        "key": "./local-action",
        "path": "local-action",
        "ref": "",
        "local": True,
        "docker": False,
        "repo": "test/repo",
    }
    assert result == expected


def test_decompose_action_ref_docker_action():
    action_path = "docker://alpine:3.8"
    repo_name = "test/repo"
    result = decompose_action_ref(action_path, repo_name)
    expected = {
        "key": "docker://alpine:3.8",
        "path": "docker://alpine:3.8",
        "ref": "",
        "local": False,
        "docker": True,
    }
    assert result == expected


def test_decompose_action_ref_standard_action_no_ref():
    action_path = "test/repo"
    repo_name = "test/repo"
    result = decompose_action_ref(action_path, repo_name)
    expected = {
        "key": "test/repo",
        "path": "",
        "ref": "",
        "local": False,
        "docker": False,
        "repo": "test/repo",
    }
    assert result == expected


def test_decompose_action_ref_nested_action_path():
    action_path = "actions/checkout/path/to/action@v2"
    repo_name = "test/repo"
    result = decompose_action_ref(action_path, repo_name)
    expected = {
        "key": "actions/checkout/path/to/action@v2",
        "path": "path/to/action",
        "ref": "v2",
        "local": False,
        "docker": False,
        "repo": "actions/checkout",
    }
    assert result == expected


@patch("gatox.workflow_parser.utility.ConfigurationManager")
def test_check_sinks_with_sink_in_script(MockConfigManager):
    MockConfigManager().WORKFLOW_PARSING = {
        "SINKS": ["dangerous_function"],
        "SINKS_START": ["start_dangerous"],
    }
    script = "This script calls dangerous_function"
    assert check_sinks(script) == True


@patch("gatox.workflow_parser.utility.ConfigurationManager")
def test_check_sinks_with_sink_start_in_script(MockConfigManager):
    MockConfigManager().WORKFLOW_PARSING = {
        "SINKS": ["dangerous_function"],
        "SINKS_START": ["start_dangerous"],
    }
    script = "start_dangerous_function()"
    assert check_sinks(script) == True


@patch("gatox.workflow_parser.utility.ConfigurationManager")
def test_check_sinks_without_sink_in_script(MockConfigManager):
    MockConfigManager().WORKFLOW_PARSING = {
        "SINKS": ["dangerous_function"],
        "SINKS_START": ["start_dangerous"],
    }
    script = "This script is safe"
    assert check_sinks(script) == False


def test_parse_github_path_with_ref():
    path = "owner/repo/path/to/file@branch"
    expected_result = ("owner/repo", "path/to/file", "branch")
    assert parse_github_path(path) == expected_result


def test_parse_github_path_without_ref():
    path = "owner/repo/path/to/file"
    expected_result = ("owner/repo", "path/to/file", "main")
    assert parse_github_path(path) == expected_result


def test_parse_github_path_with_only_repo():
    path = "owner/repo"
    expected_result = ("owner/repo", "", "main")
    assert parse_github_path(path) == expected_result


def test_parse_github_path_with_empty_path():
    path = ""
    expected_result = ("", "", "main")
    assert parse_github_path(path) == expected_result


def test_check_sus():
    item = "needs.get_ref.outputs.head-ref"

    assert check_sus(item) == True


def test_check_sus_false():
    item = "needs.get_permission.allowed"

    assert check_sus(item) == False


def test_unsafe():
    item = "github.event.pull_request.title"
    assert checkUnsafe(item) == True


def test_safe():
    item = "github.event.pull_request.url"
    assert checkUnsafe(item) == False


def test_process_runner():
    runner = "custom-mac-m1"
    assert process_runner(runner) == True


def test_rocess_runner_gh_large():
    runner = ["macos-13-xl"]
    assert process_runner(runner) == False


def test_rocess_runner_gh_large1():
    runner = ["ubuntu-24.04"]
    assert process_runner(runner) == False


def test_process_runner_list():
    runner = ["custom-mac-m1", "x64"]
    assert process_runner(runner) == True


def test_process_runner_gh():
    runner = "ubuntu-latest"
    assert process_runner(runner) == False



"""
Copyright 2025, Adnan Khan

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from concurrent.futures import ThreadPoolExecutor, wait, ALL_COMPLETED

from gatox.models.repository import Repository
from gatox.workflow_graph.graph_builder import WorkflowGraphBuilder
from gatox.git.git import Git


class IngestNonDefault:
    """Class to handle ingesting non-default branches with a thread pool."""

    _executor = None
    _api = None
    _futures = []

    @classmethod
    def ingest(cls, repo: Repository, api):
        """
        Enqueue a repository ingest task into the worker pool.
        If the pool and API haven't been initialized, initialize them.
        """
        if cls._executor is None:
            cls._executor = ThreadPoolExecutor(max_workers=8)
        if cls._api is None:
            cls._api = api

        # Submit a repository processing job
        future = cls._executor.submit(cls._process_repo, repo)
        cls._futures.append(future)
        return future

    @classmethod
    def _process_repo(cls, repo: Repository):
        """Actual processing of a repository's non-default workflows."""
        git_client = Git(cls._api.pat, repo.name)
        workflows = git_client.get_non_default()
        for workflow in workflows:
            WorkflowGraphBuilder().build_graph_from_yaml(workflow, repo)

    @classmethod
    def pool_empty(cls):
        """
        Returns a Future that completes once all currently-enqueued tasks finish.
        This method blocks until all tasks are done, then returns a completed Future.
        """
        if cls._executor is None:
            # No work was ever queued
            from concurrent.futures import Future

            f = Future()
            f.set_result(None)
            return f

        done, not_done = wait(cls._futures, return_when=ALL_COMPLETED)
        # Create a new completed Future to return
        from concurrent.futures import Future

        final_future = Future()
        final_future.set_result(None)
        return final_future



from gatox.cli.output import Fore, Style, Output
from gatox.util.arg_utils import StringType, ReadableFile, WriteableDir, WritablePath


def configure_parser_enumerate(parser):
    """Helper method to add arguments to the enumeration subparser.

    Args:
        parser: sub parser to add arguments to.
    """

    parser.add_argument(
        "--target",
        "-t",
        help="Target an organization to enumerate for self-hosted runners.",
        metavar=f"{Fore.RED}ORGANIZATION{Style.RESET_ALL}",
        type=StringType(39),
    )

    parser.add_argument(
        "--repository",
        "-r",
        help="Target a single repository in org/repo format to enumerate for\n"
        "self-hosted runners.",
        metavar=f"{Fore.RED}ORG/REPO_NAME{Style.RESET_ALL}",
        type=StringType(79, regex=r"[A-Za-z0-9-_.]+\/[A-Za-z0-9-_.]+"),
    )

    parser.add_argument(
        "--repositories",
        "-R",
        help="A text file containing repositories in org/repo format to\n"
        "enumerate for self-hosted runners.",
        metavar=f"{Fore.RED}PATH/TO/FILE.txt{Style.RESET_ALL}",
        type=ReadableFile(),
    )

    parser.add_argument(
        "--self-enumeration",
        "-s",
        help=(
            "Enumerate the configured token's access and all repositories or\n"
            "organizations the user has write access to."
        ),
        action="store_true",
    )

    parser.add_argument(
        "--validate",
        "-v",
        help=("Validate if the token is valid and print organization memberships."),
        action="store_true",
    )

    parser.add_argument(
        "--output-yaml",
        "-o",
        help=(
            "Directory to save gathered workflow yml files to. Will be\n"
            f"created in the following format: {Fore.GREEN}"
            f"org/repo/workflow.yml{Style.RESET_ALL}"
        ),
        metavar="DIR",
        type=WriteableDir(),
    )

    parser.add_argument(
        "--skip-runners",
        "-sr",
        help=(
            f"Do {Output.bright('NOT')} enumerate runners via run-log analysis, this will\n"
            "speed up the enumeration, but will miss self-hosted runners for\n"
            "non-admin users."
        ),
        action="store_true",
    )

    parser.add_argument(
        "--machine",
        help=(
            "Run with a GitHub App token, which will allow running single repository\n"
            " enumeration with server-to-server or user-to-server tokens."
        ),
        action="store_true",
    )

    parser.add_argument(
        "--ignore-workflow-run",
        help=(
            "Ignore the `workflow_run` trigger when enumerating repositories.\n"
            "This is useful if you know the organization requires approval for all\n"
            "fork pull requests."
        ),
        action="store_true",
        default=False,
    )

    parser.add_argument(
        "--output-json",
        "-oJ",
        help=("Save enumeration output to JSON file."),
        metavar="JSON_FILE",
        type=StringType(256),
    )

    parser.add_argument(
        "--deep-dive",
        "-dd",
        help=(
            "Perform deep dive static analysis, which includes analyzing non-default branches for Pwn Request vulnerabilities.\n"
            "Git is required on the PATH for this feature."
        ),
        action="store_true",
    )

    parser.add_argument(
        "--cache-restore-file",
        help=(
            "Path to JSON file containing saved reusable action files. This will reduce the need for frequent API requests."
        ),
        metavar="JSON_FILE",
        type=ReadableFile(),
    )

    parser.add_argument(
        "--cache-save-file",
        help=(
            "Path to JSON file to save cache to after executing. Can be the same as the restore file, in which case it will over-write it."
        ),
        metavar="JSON_FILE",
        type=WritablePath(),
    )



"""
Copyright 2025, Adnan Khan

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from enum import Enum


class Complexity(str, Enum):
    """
    Represents the complexity level of a workflow vulnerability.

    Values:
        ZERO_CLICK: Exploit requires no user interaction
        TOCTOU: Time-of-check to time-of-use vulnerability
        BROKEN_ACCESS: Broken access control vulnerability
        DEFAULT_DEPENDENT: Vulnerability depends on default configuration
    """

    ZERO_CLICK = "No Interaction"
    PREVIOUS_CONTRIBUTOR = "Previous Contributor"
    FOLLOW_UP = "Persistent Approval Gated"
    TOCTOU = "Time-of-Check to Time-of-Use"
    BROKEN_ACCESS = "Broken Access Control"

    def __str__(self) -> str:
        return str(self.value)

    def __repr__(self) -> str:
        return str(self.value)

    def explain(self) -> str:
        if self.value == Complexity.ZERO_CLICK:
            return "Exploit requires no user interaction, you must still confirm there are no custom permission checks that would prevent the attack."
        elif self.value == Complexity.PREVIOUS_CONTRIBUTOR:
            return "Exploit requires a previous contributor to the repository, and the repository must use the default pull-request approval setting."
        elif self.value == Complexity.FOLLOW_UP:
            return "Exploit requires a maintainer to perform some state changing action, such as labeling a PR, at that point the attacker can follow up with their payload."
        elif self.value == Complexity.TOCTOU:
            return "Exploit requires updating pull request quickly after the maintainer performs an approval action, make sure the approval action runs on forks for this to be feasible."
        elif self.value == Complexity.BROKEN_ACCESS:
            return "Exploit requires the attacker to have some access, but the access control mechanism is not properly implemented."
